from pathlib import Path

from exponent.core.remote_execution import files
from exponent.core.remote_execution.types import (
    CommandRequest,
    CommandResponse,
)
from exponent.core.types.generated.command_request_data import (
    FileReadCommandRequestData,
)


async def execute_command(
    request: CommandRequest,
    working_directory: str,
) -> CommandResponse:
    try:
        match request:
            case CommandRequest(
                correlation_id=correlation_id,
                data=FileReadCommandRequestData(file_path=file_path),
            ):
                path = Path(working_directory, file_path)
                content, _ = files.get_file_content(path)

                return CommandResponse(
                    content=content,
                    correlation_id=correlation_id,
                )

            case _:
                raise ValueError(f"Unknown command request: {request}")
    except Exception as e:  # noqa: BLE001 - TODO (Josh): Specialize errors for execution
        return CommandResponse(
            content="An error occurred during command execution: " + str(e),
            correlation_id=request.correlation_id,
        )



from collections.abc import AsyncGenerator
from typing import Any

from gql import Client, gql
from gql.transport.httpx import HTTPXAsyncTransport
from gql.transport.websockets import WebsocketsTransport


class GraphQLClient:
    def __init__(self, api_key: str, base_api_url: str):
        graphql_url = f"{base_api_url}/graphql"
        websocket_url = f"{base_api_url}/graphql_ws".replace("https", "wss").replace(
            "http", "ws"
        )

        self.transport = HTTPXAsyncTransport(
            url=graphql_url,
            headers={"API-KEY": api_key},
        )

        self.ws_transport = WebsocketsTransport(
            url=websocket_url,
            init_payload={"apiKey": api_key},
        )

    async def query(
        self,
        query_str: str,
        vars: dict[str, Any] | None = None,
        op_name: str | None = None,
    ) -> Any:
        async with Client(
            transport=self.transport,
            fetch_schema_from_transport=True,
        ) as session:
            # Execute single query
            query = gql(query_str)
            result = await session.execute(query, vars, op_name)
            return result

    async def subscribe(
        self,
        subscription_str: str,
        vars: dict[str, Any] | None = None,
    ) -> AsyncGenerator[dict[str, Any], None]:
        async with Client(
            transport=self.ws_transport,
        ) as session:
            # Execute subscription
            subscription = gql(subscription_str)
            async for result in session.subscribe(subscription, vars):
                yield result



GET_CLOUD_CHAT_STATE_QUERY = """
query GetCloudChatState($chatUuid: String!) {
    cloudChatState(chatUuid: $chatUuid) {
        __typename
        ... on CloudConnectedState {
            chatUuid
            connectedState
            lastConnectedAt
            systemInfo {
                name
                cwd
                shell
                os
                git {
                    branch
                    remote
                }
            }
        }
        ... on UnauthenticatedError {
            message
        }
        ... on ChatNotFoundError {
            message
        }
    }
}
"""



import os
from importlib.metadata import Distribution, PackageNotFoundError

import click


def file_relative_path(dunderfile: str, relative_path: str) -> str:
    """Get a path relative to the currently executing Python file."""
    return os.path.join(os.path.dirname(dunderfile), relative_path)


def get_version() -> str:
    try:
        return Distribution.from_name("exponent-run").version
    except PackageNotFoundError as e:
        click.echo(f"Error reading version: {e}", err=True)
        return "unknown"



import asyncio
import sys
import time

import click

from exponent.commands.common import (
    create_chat,
    inside_ssh_session,
    redirect_to_login,
    start_client,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.commands.utils import (
    ConnectionTracker,
    Spinner,
    launch_exponent_browser,
    print_exponent_message,
)
from exponent.core.config import Settings

try:
    # this is an optional dependency for python <3.11
    from async_timeout import timeout
except ImportError:  # pragma: no cover
    from asyncio import timeout  # type: ignore


@exponent_cli_group()
def run_cli() -> None:
    """Run AI-powered chat sessions."""
    pass


@run_cli.command()
@click.option(
    "--chat-id",
    help="ID of an existing chat session to reconnect",
    required=False,
)
@click.option(
    "--prompt",
    help="Start a chat with a given prompt.",
)
@click.option(
    "--benchmark",
    is_flag=True,
    help="Enable benchmarking mode",
)
@use_settings
def run(
    settings: Settings,
    chat_id: str | None = None,
    prompt: str | None = None,
    benchmark: bool = False,
) -> None:
    """Start or reconnect to an Exponent session."""
    if not settings.api_key:
        redirect_to_login(settings)
        return

    start_ts = time.time()
    api_key = settings.api_key
    base_url = settings.base_url
    base_api_url = settings.base_api_url
    loop = asyncio.get_event_loop()
    chat_uuid = chat_id or loop.run_until_complete(create_chat(api_key, base_api_url))

    if chat_uuid is None:
        sys.exit(1)

    if (
        (not benchmark)
        and (not prompt)
        and (not inside_ssh_session())
        # If the user specified a chat ID, they probably don't want to re-launch the chat
        and (not chat_id)
    ):
        # Open the chat in the browser
        launch_exponent_browser(settings.environment, base_url, chat_uuid)

    print_exponent_message(base_url, chat_uuid)
    print()

    connection_tracker = ConnectionTracker()

    client_fut = loop.create_task(
        start_client(
            api_key,
            base_api_url,
            chat_uuid,
            prompt,
            benchmark,
            connection_tracker,
        )
    )

    conn_fut = loop.create_task(handle_connection_changes(connection_tracker, start_ts))

    _, pending = loop.run_until_complete(
        asyncio.wait({client_fut, conn_fut}, return_when=asyncio.FIRST_COMPLETED)
    )

    print("Disconnected upon user request, shutting down...")

    for task in pending:
        task.cancel()


async def handle_connection_changes(
    connection_tracker: ConnectionTracker, start_ts: float
) -> None:
    try:
        async with timeout(5):
            assert await connection_tracker.next_change()
            print(ready_message(start_ts))
    except TimeoutError:
        spinner = Spinner("Connecting...")
        spinner.show()
        assert await connection_tracker.next_change()
        spinner.hide()
        print(ready_message(start_ts))

    while True:
        assert not await connection_tracker.next_change()

        print("Disconnected...", end="")
        await asyncio.sleep(1)
        spinner = Spinner("Reconnecting...")
        spinner.show()
        assert await connection_tracker.next_change()
        spinner.hide()
        print("\x1b[1;32m✓ Reconnected", end="")
        sys.stdout.flush()
        await asyncio.sleep(1)
        print("\r\x1b[0m\x1b[2K", end="")
        sys.stdout.flush()


def ready_message(start_ts: float) -> str:
    elapsed = round(time.time() - start_ts, 2)
    return f"\x1b[32m✓\x1b[0m Ready in {elapsed}s"



from __future__ import annotations

import asyncio
import json
import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager
from typing import TypeVar

import websockets.client
import websockets.exceptions
from httpx import (
    AsyncClient,
    codes as http_status,
)
from pydantic import BaseModel

from exponent.commands.utils import ConnectionTracker
from exponent.core.remote_execution import files, git, system_context
from exponent.core.remote_execution.code_execution import execute_code
from exponent.core.remote_execution.command_execution import execute_command
from exponent.core.remote_execution.file_write import execute_file_write
from exponent.core.remote_execution.session import (
    RemoteExecutionClientSession,
    get_session,
)
from exponent.core.remote_execution.types import (
    CLIConnectedState,
    CodeExecutionRequest,
    CommandRequest,
    CreateChatResponse,
    ExecutionEndResponse,
    FileWriteRequest,
    GetAllTrackedFilesRequest,
    GetFileAttachmentRequest,
    GetFileAttachmentsRequest,
    GetMatchingFilesRequest,
    HeartbeatInfo,
    ListFilesRequest,
    RemoteExecutionRequestType,
    RemoteExecutionResponseType,
    StartChatRequest,
    StartChatResponse,
    SystemContextRequest,
    UseToolsConfig,
)
from exponent.core.remote_execution.utils import (
    assert_unreachable,
    deserialize_api_response,
    deserialize_request_data,
    serialize_message,
)

logger = logging.getLogger(__name__)


TModel = TypeVar("TModel", bound=BaseModel)


class RemoteExecutionClient:
    def __init__(
        self,
        session: RemoteExecutionClientSession,
    ):
        self.current_session = session

        self.file_cache: files.FileCache = files.FileCache(session.working_directory)

    @property
    def working_directory(self) -> str:
        return self.current_session.working_directory

    @property
    def api_client(self) -> AsyncClient:
        return self.current_session.api_client

    async def run_connection(  # noqa: PLR0915,PLR0912
        self,
        chat_uuid: str,
        connection_tracker: ConnectionTracker | None = None,
    ) -> None:
        stop = False

        self.current_session.set_chat_uuid(chat_uuid)

        async for websocket in self.ws_connect(
            f"/api/ws/chat/{chat_uuid}",
        ):
            if connection_tracker is not None:
                await connection_tracker.set_connected(True)

            beats: asyncio.Queue[HeartbeatInfo] = asyncio.Queue()
            requests: asyncio.Queue[RemoteExecutionRequestType] = asyncio.Queue()
            results: asyncio.Queue[RemoteExecutionResponseType] = asyncio.Queue()

            async def beat() -> None:
                while True:
                    info = self.get_heartbeat_info()
                    await beats.put(info)
                    await asyncio.sleep(3)

            async def executor() -> None:
                while True:
                    request = await requests.get()
                    response = await self.handle_request(request)
                    await results.put(response)

            beat_task = asyncio.create_task(beat())
            executor_task = asyncio.create_task(executor())

            try:
                while True:
                    recv = asyncio.create_task(websocket.recv())
                    get_beat = asyncio.create_task(beats.get())
                    get_result = asyncio.create_task(results.get())

                    done, pending = await asyncio.wait(
                        [recv, get_beat, get_result],
                        return_when=asyncio.FIRST_COMPLETED,
                    )

                    for task in pending:
                        task.cancel()

                        try:
                            await task
                        except asyncio.CancelledError:
                            pass

                    if recv in done:
                        msg = json.loads(str(recv.result()))

                        if msg["type"] == "request":
                            data = json.dumps(msg["data"])
                            request = deserialize_request_data(data)
                            await requests.put(request)

                    elif get_beat in done:
                        info = get_beat.result()
                        data = json.loads(info.model_dump_json())
                        msg = json.dumps({"type": "heartbeat", "data": data})
                        await websocket.send(msg)

                    elif get_result in done:
                        response = get_result.result()
                        data = json.loads(serialize_message(response))
                        msg = json.dumps({"type": "result", "data": data})
                        await websocket.send(msg)

            except websockets.exceptions.ConnectionClosed as e:
                if e.rcvd is not None and e.rcvd.code == 1000:  # noqa: PLR2004
                    stop = True
            except TimeoutError:
                pass

            beat_task.cancel()
            executor_task.cancel()

            if stop:
                break

            if connection_tracker is not None:
                await connection_tracker.set_connected(False)

    async def check_remote_end_event(self, chat_uuid: str) -> bool:
        response = await self.api_client.get(
            f"/api/remote_execution/{chat_uuid}/execution_end",
        )
        execution_end_response = await deserialize_api_response(
            response, ExecutionEndResponse
        )
        return execution_end_response.execution_ended

    async def create_chat(self) -> CreateChatResponse:
        response = await self.api_client.post(
            "/api/remote_execution/create_chat",
        )
        return await deserialize_api_response(response, CreateChatResponse)

    async def start_chat(
        self, chat_uuid: str, prompt: str, use_tools_config: UseToolsConfig
    ) -> StartChatResponse:
        response = await self.api_client.post(
            "/api/remote_execution/start_chat",
            json=StartChatRequest(
                chat_uuid=chat_uuid,
                prompt=prompt,
                use_tools_config=use_tools_config,
            ).model_dump(),
            timeout=60,
        )
        return await deserialize_api_response(response, StartChatResponse)

    def get_heartbeat_info(self) -> HeartbeatInfo:
        return HeartbeatInfo(
            system_info=system_context.get_system_info(self.working_directory),
        )

    async def send_heartbeat(self, chat_uuid: str) -> CLIConnectedState:
        logger.info(f"Sending heartbeat for chat_uuid {chat_uuid}")
        heartbeat_info = self.get_heartbeat_info()
        response = await self.api_client.post(
            f"/api/remote_execution/{chat_uuid}/heartbeat",
            content=heartbeat_info.model_dump_json(),
            timeout=60,
        )
        if response.status_code != http_status.OK:
            raise Exception(
                f"Heartbeat failed with status code {response.status_code} and response {response.text}"
            )
        connected_state = await deserialize_api_response(response, CLIConnectedState)
        logger.info(f"Heartbeat response: {connected_state}")
        return connected_state

    async def handle_request(
        self, request: RemoteExecutionRequestType
    ) -> RemoteExecutionResponseType:
        response: RemoteExecutionResponseType
        if isinstance(request, CodeExecutionRequest):
            response = await execute_code(
                request, self.current_session, working_directory=self.working_directory
            )
        elif isinstance(request, FileWriteRequest):
            response = execute_file_write(
                request, working_directory=self.working_directory
            )
        elif isinstance(request, ListFilesRequest):
            response = files.list_files(request)
        elif isinstance(request, GetFileAttachmentRequest):
            response = files.get_file_attachment(request, self.working_directory)
        elif isinstance(request, GetFileAttachmentsRequest):
            response = files.get_file_attachments(request, self.working_directory)
        elif isinstance(request, GetMatchingFilesRequest):
            response = await files.get_matching_files(request, self.file_cache)
        elif isinstance(request, SystemContextRequest):
            response = system_context.get_system_context(
                request, self.working_directory
            )
        elif isinstance(request, GetAllTrackedFilesRequest):
            response = await git.get_all_tracked_files(request, self.working_directory)
        elif isinstance(request, CommandRequest):
            response = await execute_command(request, self.working_directory)
        else:
            assert_unreachable(request)
        return response

    def ws_connect(self, path: str) -> websockets.client.connect:
        base_url = (
            str(self.api_client.base_url)
            .replace("http://", "ws://")
            .replace("https://", "wss://")
        )

        url = f"{base_url}{path}"
        headers = {"api-key": self.api_client.headers["api-key"]}

        return websockets.client.connect(
            url, extra_headers=headers, timeout=10, ping_timeout=10
        )

    @staticmethod
    @asynccontextmanager
    async def session(
        api_key: str, base_url: str, working_directory: str
    ) -> AsyncGenerator[RemoteExecutionClient, None]:
        async with get_session(working_directory, base_url, api_key) as session:
            yield RemoteExecutionClient(session)



# Auto-generated, do not edit directly. Run `make generate_command_data` to update.

from enum import Enum
from typing import Literal

from pydantic import BaseModel


class CommandRequestType(str, Enum):
    FILE_READ = "file_read"
    FILE_OPEN = "file_open"


class CommandRequestData(BaseModel):
    pass


class FileReadCommandRequestData(CommandRequestData):
    type: Literal[CommandRequestType.FILE_READ] = CommandRequestType.FILE_READ

    file_path: str
    language: str


class FileOpenCommandRequestData(CommandRequestData):
    type: Literal[CommandRequestType.FILE_OPEN] = CommandRequestType.FILE_OPEN

    file_path: str
    language: str


CommandRequestDataType = FileReadCommandRequestData | FileOpenCommandRequestData



import asyncio
import logging
import os
import sys
from collections.abc import Coroutine
from typing import Any

import click
import httpx
from dotenv import load_dotenv

from exponent.commands.utils import ConnectionTracker
from exponent.core.config import (
    Settings,
    get_settings,
)
from exponent.core.remote_execution.client import RemoteExecutionClient
from exponent.core.remote_execution.exceptions import ExponentError
from exponent.core.remote_execution.types import UseToolsConfig

load_dotenv()


def set_log_level() -> None:
    settings = get_settings()
    logging.basicConfig(level=getattr(logging, settings.log_level), stream=sys.stdout)


def redirect_to_login(settings: Settings, cause: str = "detected") -> None:
    if inside_ssh_session():
        click.echo(f"No API Key {cause}, run 'exponent login --key <API-KEY>'")
    else:
        click.echo("No API Key detected, redirecting to login...")
        click.launch(f"{settings.base_url}/")


def inside_ssh_session() -> bool:
    return (os.environ.get("SSH_TTY") or os.environ.get("SSH_TTY")) is not None


def run_until_complete(coro: Coroutine[Any, Any, Any]) -> Any:
    loop = asyncio.get_event_loop()
    task = loop.create_task(coro)

    try:
        loop.run_until_complete(task)
    except KeyboardInterrupt:
        task.cancel()
        try:
            loop.run_until_complete(task)
        except asyncio.CancelledError:
            pass
    except ExponentError as e:
        click.secho(f"Encountered error: {e}", fg="red")
        click.secho(
            "The Exponent team has been notified, "
            "please try again and reach out if the problem persists.",
            fg="yellow",
        )
        sys.exit(1)


async def benchmark_mode_exit(client: RemoteExecutionClient, chat_uuid: str) -> None:
    while True:
        await asyncio.sleep(5)
        if await client.check_remote_end_event(chat_uuid):
            sys.exit(0)


async def run_client_connection(
    client: RemoteExecutionClient,
    chat_uuid: str,
    connection_tracker: ConnectionTracker | None = None,
) -> None:
    await client.run_connection(chat_uuid, connection_tracker)


async def start_chat(
    client: RemoteExecutionClient,
    chat_uuid: str,
    prompt: str,
    use_tools_config: UseToolsConfig,
) -> None:
    click.secho("Starting chat...")
    await client.start_chat(chat_uuid, prompt, use_tools_config=use_tools_config)
    click.secho("Chat started. Open the link to join the chat.")


async def start_client(
    api_key: str,
    base_api_url: str,
    chat_uuid: str,
    prompt: str | None = None,
    benchmark: bool = False,
    connection_tracker: ConnectionTracker | None = None,
) -> None:
    if benchmark is True and prompt is None:
        click.secho("Error: Benchmark mode requires a prompt.", fg="red")
        return

    current_working_directory = os.getcwd()

    use_tools_config = UseToolsConfig()
    async with RemoteExecutionClient.session(
        api_key, base_api_url, current_working_directory
    ) as client:
        if benchmark:
            assert prompt is not None
            await asyncio.gather(
                start_chat(
                    client, chat_uuid, prompt, use_tools_config=use_tools_config
                ),
                run_client_connection(client, chat_uuid, connection_tracker),
                benchmark_mode_exit(client, chat_uuid),
            )
        elif prompt:
            await asyncio.gather(
                start_chat(
                    client, chat_uuid, prompt, use_tools_config=use_tools_config
                ),
                run_client_connection(client, chat_uuid, connection_tracker),
            )
        else:
            await run_client_connection(client, chat_uuid, connection_tracker)


# Helper functions
async def create_chat(api_key: str, base_api_url: str) -> str | None:
    try:
        async with RemoteExecutionClient.session(
            api_key, base_api_url, os.getcwd()
        ) as client:
            chat = await client.create_chat()
            return chat.chat_uuid
    except (httpx.ConnectError, ExponentError) as e:
        click.secho(f"Error: {e}", fg="red")
        return None



# Auto-generated, do not edit directly. Run `make generate_strategy_info` to update.

import enum

from pydantic import BaseModel


class StrategyName(str, enum.Enum):
    FULL_FILE_REWRITE = "FULL_FILE_REWRITE"
    NATURAL_EDIT = "NATURAL_EDIT"
    UDIFF = "UDIFF"
    SEARCH_REPLACE = "SEARCH_REPLACE"
    OPEN_FILE_SEARCH_REPLACE = "OPEN_FILE_SEARCH_REPLACE"
    FUNCTION_CALLING = "FUNCTION_CALLING"
    RAW_GPT = "RAW_GPT"
    STRUCT_OUT = "STRUCT_OUT"


class StrategyInfo(BaseModel):
    strategy_name: StrategyName
    display_name: str
    description: str
    disabled: bool
    display_order: int


STRATEGY_INFO_LIST: list[StrategyInfo] = [
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT,
        display_name="Natural Edit",
        description="A natural file editing strategy.",
        disabled=False,
        display_order=0,
    ),
    StrategyInfo(
        strategy_name=StrategyName.FULL_FILE_REWRITE,
        display_name="Full File Rewrites",
        description="Rewrites the full file every time. Use this if your files are generally less than 300 lines.",
        disabled=False,
        display_order=1,
    ),
    StrategyInfo(
        strategy_name=StrategyName.UDIFF,
        display_name="Unified Diff",
        description="Generates diffs to edit files",
        disabled=False,
        display_order=2,
    ),
    StrategyInfo(
        strategy_name=StrategyName.SEARCH_REPLACE,
        display_name="Search and Replace",
        description="Replaces chunks of code with new version of code. Recommended strategy for larger more complex files.",
        disabled=False,
        display_order=3,
    ),
    StrategyInfo(
        strategy_name=StrategyName.OPEN_FILE_SEARCH_REPLACE,
        display_name="Open + Search/Replace",
        description="Uses the Open File command to open files and Search/Replace to edit.",
        disabled=False,
        display_order=4,
    ),
    StrategyInfo(
        strategy_name=StrategyName.FUNCTION_CALLING,
        display_name="Function Calling",
        description="No description",
        disabled=True,
        display_order=99,
    ),
    StrategyInfo(
        strategy_name=StrategyName.RAW_GPT,
        display_name="Raw GPT",
        description="No description",
        disabled=True,
        display_order=99,
    ),
    StrategyInfo(
        strategy_name=StrategyName.STRUCT_OUT,
        display_name="Structured Output",
        description="Uses GPTs strict json_schema format.",
        disabled=True,
        display_order=99,
    ),
]


ENABLED_STRATEGY_INFO_LIST: list[StrategyInfo] = [
    StrategyInfo(
        strategy_name=StrategyName.NATURAL_EDIT,
        display_name="Natural Edit",
        description="A natural file editing strategy.",
        disabled=False,
        display_order=0,
    ),
    StrategyInfo(
        strategy_name=StrategyName.FULL_FILE_REWRITE,
        display_name="Full File Rewrites",
        description="Rewrites the full file every time. Use this if your files are generally less than 300 lines.",
        disabled=False,
        display_order=1,
    ),
    StrategyInfo(
        strategy_name=StrategyName.UDIFF,
        display_name="Unified Diff",
        description="Generates diffs to edit files",
        disabled=False,
        display_order=2,
    ),
    StrategyInfo(
        strategy_name=StrategyName.SEARCH_REPLACE,
        display_name="Search and Replace",
        description="Replaces chunks of code with new version of code. Recommended strategy for larger more complex files.",
        disabled=False,
        display_order=3,
    ),
    StrategyInfo(
        strategy_name=StrategyName.OPEN_FILE_SEARCH_REPLACE,
        display_name="Open + Search/Replace",
        description="Uses the Open File command to open files and Search/Replace to edit.",
        disabled=False,
        display_order=4,
    ),
]



import json

import click

from exponent.commands.common import (
    redirect_to_login,
    run_until_complete,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.core.config import Settings, get_settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.cloud_config_queries import (
    CREATE_CLOUD_CONFIG_MUTATION,
    GET_CLOUD_CONFIGS_QUERY,
    UPDATE_CLOUD_CONFIG_MUTATION,
)
from exponent.core.graphql.get_chats_query import GET_CHATS_QUERY
from exponent.core.graphql.github_config_queries import CREATE_GITHUB_CONFIG_MUTATION
from exponent.core.graphql.subscriptions import AUTHENTICATED_USER_SUBSCRIPTION


@exponent_cli_group()
def config_cli() -> None:
    """Manage Exponent configuration settings."""
    pass


@config_cli.command()
def config() -> None:
    """Display current Exponent configuration."""
    config_file_settings = get_settings().get_config_file_settings()

    click.secho(
        json.dumps(config_file_settings, indent=2),
        fg="green",
    )


@config_cli.command()
@click.option("--key", help="Your Exponent API Key")
@use_settings
def login(settings: Settings, key: str) -> None:
    """Log in to Exponent."""

    if not key:
        redirect_to_login(settings, "provided")
        return

    click.echo(f"Saving API Key to {settings.config_file_path}")

    if settings.api_key and settings.api_key != key:
        click.confirm("Detected existing API Key, continue? ", default=True, abort=True)

    settings.update_api_key(key)
    settings.write_settings_to_config_file()

    click.echo("API Key saved.")


@config_cli.command(hidden=True)
@use_settings
def get_chats(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_chats_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
        )
    )


@config_cli.command(hidden=True)
@use_settings
def get_authenticated_user(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_authenticated_user_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
        )
    )


async def get_chats_task(
    api_key: str,
    base_api_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url)
    result = await graphql_client.query(GET_CHATS_QUERY)
    click.echo(result)


async def get_authenticated_user_task(
    api_key: str,
    base_api_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url)
    async for it in graphql_client.subscribe(AUTHENTICATED_USER_SUBSCRIPTION):
        click.echo(it)


@config_cli.command(hidden=True)
@use_settings
def get_cloud_configs(
    settings: Settings,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        get_cloud_configs_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
        )
    )


async def get_cloud_configs_task(
    api_key: str,
    base_api_url: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url)
    result = await graphql_client.query(GET_CLOUD_CONFIGS_QUERY)
    click.echo(result)


async def create_github_config_task(
    api_key: str,
    base_api_url: str,
    github_pat: str,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url)
    variables = {
        "githubPat": github_pat,
    }
    result = await graphql_client.query(CREATE_GITHUB_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option("--github-pat", required=True, help="Github personal access token")
@use_settings
def create_github_config(
    settings: Settings,
    github_pat: str,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        create_github_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            github_pat=github_pat,
        )
    )


@config_cli.command(hidden=True)
@click.option("--github-org-name", required=True, help="GitHub organization name")
@click.option("--github-repo-name", required=True, help="GitHub repository name")
@click.option(
    "--setup_commands",
    required=False,
    help="List of commands to set up and build your repo",
    multiple=True,
)
@use_settings
def create_cloud_config(
    settings: Settings,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        create_cloud_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            github_org_name=github_org_name,
            github_repo_name=github_repo_name,
            setup_commands=setup_commands,
        )
    )


async def create_cloud_config_task(
    api_key: str,
    base_api_url: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url)
    variables = {
        "githubOrgName": github_org_name,
        "githubRepoName": github_repo_name,
        "setupCommands": setup_commands,
    }
    result = await graphql_client.query(CREATE_CLOUD_CONFIG_MUTATION, variables)
    click.echo(result)


@config_cli.command(hidden=True)
@click.option("--cloud-config-uuid", required=True, help="Cloud config UUID")
@click.option("--github-org-name", required=True, help="GitHub organization name")
@click.option("--github-repo-name", required=True, help="GitHub repository name")
@click.option(
    "--setup_commands",
    required=False,
    help="List of commands to set up and build your repo",
    multiple=True,
)
@use_settings
def update_cloud_config(
    settings: Settings,
    cloud_config_uuid: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        update_cloud_config_task(
            api_key=settings.api_key,
            base_api_url=settings.base_api_url,
            cloud_config_uuid=cloud_config_uuid,
            github_org_name=github_org_name,
            github_repo_name=github_repo_name,
            setup_commands=setup_commands,
        )
    )


async def update_cloud_config_task(
    api_key: str,
    base_api_url: str,
    cloud_config_uuid: str,
    github_org_name: str,
    github_repo_name: str,
    setup_commands: list[str] | None,
) -> None:
    graphql_client = GraphQLClient(api_key, base_api_url)
    variables = {
        "cloudConfigUuid": cloud_config_uuid,
        "githubOrgName": github_org_name,
        "githubRepoName": github_repo_name,
        "setupCommands": setup_commands,
    }

    result = await graphql_client.query(UPDATE_CLOUD_CONFIG_MUTATION, variables)
    click.echo(result)



import click

from exponent.commands.types import exponent_cli_group
from exponent.commands.utils import (
    check_exponent_version,
    get_exponent_version_for_update,
)
from exponent.version import get_version


@exponent_cli_group()
def upgrade_cli() -> None:
    """Manage Exponent version upgrades."""
    pass


@upgrade_cli.command()
@click.option(
    "--force", is_flag=True, help="Force upgrade even if no new version is available."
)
def upgrade(force: bool = False) -> None:
    """Upgrade Exponent to the latest version."""
    current_version = get_version()
    new_version = get_exponent_version_for_update()

    if new_version or force:
        click.echo(f"Current version: {current_version}")
        if new_version:
            click.echo(f"New version available: {new_version}")
        check_exponent_version()
    else:
        click.echo("Exponent is already up to date.")



HALT_CHAT_STREAM_MUTATION: str = """
  mutation HaltChatStream($chatUuid: String!) {
    haltChatStream(chatUuid: $chatUuid) {
      __typename
    }
  }
"""


MOVE_CHAT_TO_CLOUD_MUTATION: str = """
  mutation MoveChatToCloud($chatUuid: String!) {
    moveChatToCloud(chatUuid: $chatUuid) {
      ... on Chat {
        chatUuid
        name
      }
      ... on UnauthenticatedError {
        message
      }
      ... on ChatNotFoundError {
        message
      }
      ... on CloudSessionError {
        message
        chatUuid
      }
    }
  }
"""



import asyncio
import os
from collections.abc import Callable
from pathlib import Path
from typing import Any

from gitignore_parser import (
    IgnoreRule,
    handle_negation,
    parse_gitignore,
    rule_from_pattern,
)
from rapidfuzz import process

from exponent.core.remote_execution.types import (
    FileAttachment,
    GetFileAttachmentRequest,
    GetFileAttachmentResponse,
    GetFileAttachmentsRequest,
    GetFileAttachmentsResponse,
    GetMatchingFilesRequest,
    GetMatchingFilesResponse,
    ListFilesRequest,
    ListFilesResponse,
    RemoteFile,
)

FILE_NOT_FOUND = "File {} does not exist"


class FileCache:
    def __init__(self, working_directory: str):
        self.working_directory = working_directory
        self._cache: list[str] | None = None

    async def get_files(self) -> list[str]:
        if self._cache is None:
            self._cache = await file_walk(self.working_directory)
        return self._cache


def list_files(list_files_request: ListFilesRequest) -> ListFilesResponse:
    filenames = os.listdir(list_files_request.directory)
    return ListFilesResponse(
        files=[
            RemoteFile(
                file_path=filename,
                working_directory=list_files_request.directory,
            )
            for filename in filenames
        ],
        correlation_id=list_files_request.correlation_id,
    )


def read_file(
    file_path: str | Path,
) -> str | None:
    if not os.path.exists(file_path):
        return None
    with open(file_path) as f:
        return f.read()


def get_file_content(
    absolute_path: str | Path,
) -> tuple[str, bool]:
    exists = True
    content = read_file(absolute_path)

    if content is None:
        exists = False
        content = FILE_NOT_FOUND.format(absolute_path)

    return content, exists


def get_file_attachments(
    get_file_attachments_request: GetFileAttachmentsRequest,
    client_working_directory: str,
) -> GetFileAttachmentsResponse:
    file_attachment_responses = []

    for file in get_file_attachments_request.files:
        absolute_path = file.resolve(client_working_directory)
        content, _ = get_file_content(absolute_path)

        file_attachment_responses.append(
            FileAttachment(
                file=file,
                content=content,
            )
        )

    return GetFileAttachmentsResponse(
        file_attachments=file_attachment_responses,
        correlation_id=get_file_attachments_request.correlation_id,
    )


def get_file_attachment(
    get_file_attachment_request: GetFileAttachmentRequest, client_working_directory: str
) -> GetFileAttachmentResponse:
    file = get_file_attachment_request.file
    absolute_path = file.resolve(client_working_directory)

    content, exists = get_file_content(absolute_path)

    return GetFileAttachmentResponse(
        content=content,
        exists=exists,
        file=file,
        correlation_id=get_file_attachment_request.correlation_id,
    )


def _parse_gitignore(directory: str) -> Any:
    gitignore_path = os.path.join(directory, ".gitignore")
    if os.path.isfile(gitignore_path):
        return parse_gitignore(gitignore_path)
    return None


def _parse_ignore_extra(
    working_directory: str, ignore_extra: list[str]
) -> Callable[[str], bool]:
    base_path = Path(working_directory).resolve()
    rules: list[IgnoreRule] = []
    for pattern in ignore_extra:
        if (rule := rule_from_pattern(pattern, base_path=base_path)) is not None:
            rules.append(rule)

    def rule_handler(file_path: str) -> bool:
        nonlocal rules
        resolved_path = str(Path(file_path).resolve())
        return bool(handle_negation(resolved_path, rules))

    return rule_handler


def _or(a: Callable[[str], bool], b: Callable[[str], bool]) -> Callable[[str], bool]:
    def or_handler(file_path: str) -> bool:
        return a(file_path) or b(file_path)

    return or_handler


def _get_ignored_checker(
    dir_path: str, existing_ignores: dict[str, Any]
) -> Callable[[str], bool] | None:
    new_ignore = _parse_gitignore(dir_path)
    existing_ignore = existing_ignores.get(dir_path)
    if existing_ignore and new_ignore:
        return _or(new_ignore, existing_ignore)
    return new_ignore or existing_ignore


async def file_walk(
    working_directory: str,
    ignore_extra: list[str] | None = None,
    max_files: int = 10_000,
) -> list[str]:
    working_path = Path(working_directory).resolve()

    ignored_checkers = {}
    if ignore_extra:
        # Optional extra gitignore patterns
        ignored_checkers[str(working_path)] = _parse_ignore_extra(
            working_directory, ignore_extra
        )

    all_files = []
    for dirpath, dirnames, filenames in os.walk(working_directory):
        # Update or add new .gitignore rules when a .gitignore file is encountered
        dirpath_resolved = str(Path(dirpath).resolve())
        new_ignore = _get_ignored_checker(dirpath_resolved, ignored_checkers)
        if new_ignore:
            ignored_checkers[dirpath_resolved] = new_ignore

        # Check each file in the current directory
        for filename in filenames:
            file_path = os.path.join(dirpath_resolved, filename)
            file_path = str(Path(file_path).resolve())
            # Check against all applicable .gitignore rules
            ignored = any(
                ignored_checkers[d](file_path)
                for d in ignored_checkers
                if is_subpath(file_path, d)
            )
            if not ignored and (relpath := get_relpath(file_path, working_path)):
                all_files.append(relpath)

        if len(all_files) >= max_files:
            break

        # Update directory list in place to skip ignored directories
        dirnames[:] = [
            d
            for d in dirnames
            if not any(
                ignored_checkers[dp](os.path.join(dirpath_resolved, d))
                for dp in ignored_checkers
                if is_subpath(os.path.join(dirpath_resolved, d), dp)
            )
            and ".git" not in d
        ]

        await asyncio.sleep(0)

    return all_files


async def get_matching_files(
    search_term: GetMatchingFilesRequest,
    file_cache: FileCache,
) -> GetMatchingFilesResponse:
    MAX_MATCHING_FILES = 10

    # Use rapidfuzz to find the best matching files
    matching_files = process.extract(
        search_term.search_term,
        await file_cache.get_files(),
        limit=MAX_MATCHING_FILES,
        score_cutoff=0,
    )

    directory = file_cache.working_directory
    files: list[RemoteFile] = []
    for file, _, _ in matching_files:
        files.append(
            RemoteFile(
                file_path=file,
                working_directory=directory,
            )
        )

    return GetMatchingFilesResponse(
        files=files,
        correlation_id=search_term.correlation_id,
    )


def normalize_files(working_directory: str, file_paths: list[str]) -> list[RemoteFile]:
    working_path = Path(working_directory).resolve()
    normalized_files = []

    for file_path in file_paths:
        relative_path = Path(file_path)

        if relative_path.is_absolute():
            relative_path = relative_path.relative_to(working_path)

        normalized_files.append(
            RemoteFile(
                file_path=str(relative_path),
                working_directory=working_directory,
            )
        )

    return sorted(normalized_files)


def is_subpath(path: str, parent: str) -> bool:
    """
    Check if a path is a subpath of another path.
    """
    return os.path.commonpath([path, parent]) == parent


def get_relpath(path: str | Path, parent: str | Path) -> str | None:
    """
    Get the relative path of a file from a parent directory.
    """
    try:
        return str(Path(str(path)).relative_to(str(parent)))
    except ValueError:
        return None



import asyncio
import os
import subprocess
import sys
import threading
import time
from json import JSONDecodeError

import click
import httpx
from packaging.version import Version

from exponent.core.config import Environment, ExponentCloudConfig, Settings
from exponent.version import get_version


def print_editable_install_forced_prod_warning(settings: Settings) -> None:
    click.secho(
        "Detected local editable install, but this command only works against prod.",
        fg="red",
        bold=True,
    )
    click.secho("Using prod settings:", fg="red", bold=True)
    click.secho("- base_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
    click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_api_url}", fg=(100, 200, 255), bold=False)
    click.secho()


def print_editable_install_warning(settings: Settings) -> None:
    click.secho(
        "Detected local editable install, using local URLs", fg="yellow", bold=True
    )
    click.secho("- base_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_url}", fg=(100, 200, 255), bold=False)
    click.secho("- base_api_url=", fg="yellow", bold=True, nl=False)
    click.secho(f"{settings.base_api_url}", fg=(100, 200, 255), bold=False)
    click.secho()


def print_exponent_message(base_url: str, chat_uuid: str) -> None:
    version = get_version()
    shell = os.environ.get("SHELL")

    click.echo()
    click.secho(f"△ Exponent v{version}", fg=(180, 150, 255), bold=True)
    click.echo()
    click.echo(
        " - Link: " + click.style(f"{base_url}/chats/{chat_uuid}", fg=(100, 200, 255))
    )

    if shell is not None:
        click.echo(f" - Shell: {shell}")


def is_exponent_app_installed() -> bool:
    if sys.platform == "darwin":  # macOS
        return os.path.exists("/Applications/Exponent.app")

    # TODO: Add support for Windows and Linux
    return False


def launch_exponent_browser(
    environment: Environment, base_url: str, chat_uuid: str
) -> None:
    if is_exponent_app_installed() and environment == Environment.production:
        url = f"exponent://chats/{chat_uuid}"
    else:
        url = f"{base_url}/chats/{chat_uuid}"
    click.launch(url)


def write_template_exponent_cloud_config(file_path: str) -> None:
    exponent_cloud_config = ExponentCloudConfig(
        repo_name="your_repo_name",
        repo_specific_setup_commands=[
            "cd /home/user",
            "gh repo clone https://github.com/<org>/<repo>.git",
            "cd <repo>",
            "# Any additional setup commands",
        ],
        gh_token="ghp_your_token_here",
        runloop_api_key="ak_your_runloop_api_key_here",
    )
    with open(file_path, "w") as f:
        f.write(exponent_cloud_config.model_dump_json(indent=2))


def get_exponent_version_for_update() -> str | None:
    """
    Check if there's a newer version of Exponent available on PyPI.
    """
    installed_version = get_version()

    try:
        pypi_package_info = httpx.get("https://pypi.org/pypi/exponent-run/json").json()
    except (httpx.ConnectError, JSONDecodeError):
        return None
    pypi_version = pypi_package_info.get("info", {}).get("version")

    if not pypi_version or installed_version == "unknown":
        # If we can't get the concrete version info,
        # we can't tell if it's outdated
        return None
    elif Version(pypi_version) > Version(installed_version):
        return str(pypi_version)
    return None


def check_exponent_version() -> None:
    """
    Print a message to the user if there's a newer
    version of Exponent available on PyPI.
    """
    current_version = get_version()
    if current_version == "unknown":
        click.secho("Unable to determine current Exponent version.", fg="yellow")
        return

    new_version = get_exponent_version_for_update()
    if not new_version:
        return

    upgrade_command = ["pip", "install", "--upgrade", "exponent-run"]
    click.secho(
        f"New version available: exponent-run=={new_version} (current: {current_version})\n"
        f"Update command: `{' '.join(upgrade_command)}`",
        fg="yellow",
        bold=True,
    )
    confirmed = click.confirm("Update now?", default=True, show_default=True)
    if not confirmed:
        click.secho("Not updating.", fg="red")
        return

    upgrade_exponent(new_version)


def upgrade_exponent(new_version: str) -> None:
    """
    Upgrade Exponent to the latest version.
    """
    upgrade_command = ["pip", "install", "--upgrade", "exponent-run"]
    click.secho("Updating...", bold=True, fg="yellow")
    subprocess.check_call(upgrade_command)
    click.secho(f"Successfully upgraded Exponent to version {new_version}!", fg="green")
    invoking_command = subprocess.list2cmdline(["exponent", *sys.argv[1:]])
    click.secho(
        "\nUpdate complete! Please re-run command: ",
        fg="green",
        nl=False,
    )
    click.secho(
        invoking_command,
        bold=True,
    )
    sys.exit(0)


def start_background_event_loop() -> asyncio.AbstractEventLoop:
    def run_event_loop(loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(loop)
        loop.run_forever()

    loop = asyncio.new_event_loop()
    thread = threading.Thread(target=run_event_loop, args=(loop,), daemon=True)
    thread.start()
    return loop


def read_input(prompt: str) -> str:
    sys.stdout.write(prompt)
    sys.stdout.flush()
    return sys.stdin.readline()


class Spinner:
    def __init__(self, text: str) -> None:
        self.text = text
        self.task: asyncio.Task[None] | None = None
        self.base_time = time.time()

    def show(self) -> None:
        if self.task is not None:
            return

        async def spinner(base_time: float) -> None:
            chars = "⣷⣯⣟⡿⢿⣻⣽⣾"

            while True:
                t = time.time() - base_time
                i = round(t * 10) % len(chars)
                print(f"\r{chars[i]} {self.text}", end="")
                await asyncio.sleep(0.1)

        self.task = asyncio.get_event_loop().create_task(spinner(self.base_time))

    def hide(self) -> None:
        if self.task is None:
            return

        self.task.cancel()
        self.task = None
        print("\r\x1b[0m\x1b[2K", end="")
        sys.stdout.flush()


class ConnectionTracker:
    def __init__(self) -> None:
        self.connected = True
        self.queue: asyncio.Queue[bool] = asyncio.Queue()

    def is_connected(self) -> bool:
        while True:
            try:
                self.connected = self.queue.get_nowait()
            except asyncio.QueueEmpty:
                break

        return self.connected

    async def wait_for_reconnection(self) -> None:
        if not self.is_connected():
            assert await self.queue.get()
            self.connected = True

    async def set_connected(self, connected: bool) -> None:
        await self.queue.put(connected)

    async def next_change(self) -> bool:
        return await self.queue.get()



import logging
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

import httpx
from httpx import AsyncClient, Request, Response

from exponent.core.remote_execution.exceptions import ExponentError
from exponent.core.remote_execution.languages.python import Kernel
from exponent.core.remote_execution.utils import format_error_log

logger = logging.getLogger(__name__)


class SessionLog:
    def __init__(self) -> None:
        self.log_buffer: list[str] = []
        self.max_size = 5

    def append_log(self, log: str) -> None:
        self.log_buffer.append(log)
        self.log_buffer = self.log_buffer[-self.max_size :]

    def get_logs(self) -> list[str]:
        return self.log_buffer

    async def log_request(self, request: Request) -> None:
        self.append_log(f"Request: {request.method} {request.url}")

    async def log_response(self, response: Response) -> None:
        request = response.request
        await response.aread()
        self.append_log(
            f"Response for request: {request.method} {request.url}\n"
            f"Response: {response.status_code}, {response.text}"
        )


class RemoteExecutionClientSession:
    def __init__(self, working_directory: str, base_url: str, api_key: str):
        self.chat_uuid: str | None = None

        self.working_directory = working_directory
        self.kernel = Kernel(working_directory=working_directory)
        self.api_log = SessionLog()

        self.api_client = AsyncClient(
            base_url=base_url,
            headers={"API-KEY": api_key},
            event_hooks={
                "request": [self.api_log.log_request],
                "response": [self.api_log.log_response],
            },
        )

    def set_chat_uuid(self, chat_uuid: str) -> None:
        self.chat_uuid = chat_uuid

    async def send_exception_log(self, exc: Exception) -> None:
        error_log = format_error_log(
            exc=exc,
            chat_uuid=self.chat_uuid,
            attachment_lines=self.api_log.get_logs(),
        )

        if not error_log:
            return
        try:
            await self.api_client.post(
                "/api/remote_execution/log_error",
                content=error_log.model_dump_json(),
                timeout=60,
            )
        except httpx.ConnectError:
            logger.error("Failed to send error log")


@asynccontextmanager
async def get_session(
    working_directory: str,
    base_url: str,
    api_key: str,
) -> AsyncGenerator[RemoteExecutionClientSession, None]:
    session = RemoteExecutionClientSession(working_directory, base_url, api_key)
    try:
        yield session
    except Exception as exc:  # noqa: BLE001
        await session.send_exception_log(exc)
        raise ExponentError(str(exc))
    finally:
        session.kernel.close()
        await session.api_client.aclose()



import json
import logging
from typing import Any

from httpx import (
    AsyncClient,
    Response,
)

logger = logging.getLogger(__name__)
_timeout = 30


class RunloopClient:
    def __init__(self, api_key: str, base_url: str = "https://api.runloop.ai"):
        self.headers = {"Authorization": f"Bearer {api_key}"}
        self.base_url = base_url

    async def create_devbox(
        self,
        entrypoint: str | None | None = None,
        environment_variables: dict[str, str] | None = None,
        setup_commands: list[str] | None = None,
    ) -> dict[str, Any]:
        """
        Create a running devbox.

        Args:
            entrypoint: (Optional) When specified, the Devbox will run this script as its main executable.
                The devbox lifecycle will be bound to entrypoint, shutting down when the process is complete.

            environment_variables: (Optional) Environment variables used to configure your Devbox.

            setup_commands: (Optional) List of commands needed to set up your Devbox. Examples might include
                fetching a tool or building your dependencies. Runloop will look optimize these steps for you.
        Returns:
            Devbox instance in the form of:
            {
                "id": str,
                "status": str (provisioning, initializing, running, failure,  shutdown),
                "create_time_ms": long
            }
        """

        async with AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/v1/devboxes/",
                headers=self.headers,
                json={
                    "entrypoint": entrypoint,
                    "environment_variables": environment_variables,
                    "setup_commands": setup_commands,
                },
                # This is 30 seconds because the devbox might not be waricl
                timeout=_timeout,
            )
        return self._response_json(response)

    async def get_devbox(self, id: str) -> dict[str, Any]:
        """
        Get updated devbox.

        Args:
            id: Id of the devbox instance.
        Returns:
            Devbox instance in the form of:
            {
                "id": str,
                "status": str (provisioning, initializing, running, failure,  shutdown),
                "create_time_ms": long
            }
        """
        async with AsyncClient() as client:
            response = await client.get(
                f"{self.base_url}/v1/devboxes/{id}",
                headers=self.headers,
            )
        return self._response_json(response)

    async def shutdown_devbox(self, id: str) -> dict[str, Any]:
        """
        Shutdown devbox.

        Args:
            id: Id of the devbox instance.
        Returns:
            Updated devbox instance in the form of:
            {
                "id": str,
                "status": str (provisioning, initializing, running, failure,  shutdown),
                "create_time_ms": long
            }
        """
        async with AsyncClient() as client:
            response = await client.post(
                f"{self.base_url}/v1/devboxes/{id}/shutdown",
                headers=self.headers,
            )
        return self._response_json(response)

    async def list_devboxes(self) -> dict[str, Any]:
        """
        List previously created devboxes.

        Returns:
            A list of devbox instances in the form of:
            {
                "devboxes: [
                    "id": str,
                    "status": str (provisioning, initializing, running, failure,  shutdown),
                    "create_time_ms": long
                ]
            }
        """
        async with AsyncClient() as client:
            response = await client.get(
                f"{self.base_url}/v1/devboxes/",
                headers=self.headers,
            )
        return self._response_json(response)

    async def devbox_logs(self, id: str) -> dict[str, Any]:
        """
        Get logs from a particular devbox instance.

        Args:
            id: Id of the devbox instance.
        Returns:
            A devbox object of the following shape:
            {
                "logs: [
                    "message": str,
                    "level": str,
                    "timestamp_ms": long
                ]
            }
        """
        async with AsyncClient() as client:
            response = await client.get(
                f"{self.base_url}/v1/devboxes/{id}/logs",
                headers=self.headers,
            )
        return self._response_json(response)

    def _response_json(self, response: Response) -> dict[str, Any]:
        try:
            response_json = response.json()
        except json.JSONDecodeError as e:
            logger.error(f"Error parsing response: {e}")
            logger.error(
                f"Raw response: [status-{response.status_code}] {response.text}"
            )
            raise e

        return dict(response_json)



import asyncio
import os
import shutil

STDOUT_FD = 1
STDERR_FD = 2
MAX_TIMEOUT = 300


async def execute_shell(code: str, working_directory: str, timeout: int) -> str:
    timeout = min(timeout, MAX_TIMEOUT)

    shell_path = (
        os.environ.get("SHELL")
        or shutil.which("bash")
        or shutil.which("sh")
        or "/bin/sh"
    )

    process = await asyncio.create_subprocess_exec(
        shell_path,
        "-l",
        "-c",
        code,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        cwd=working_directory,
    )

    exit_code = None
    output: list[tuple[int, str]] = []
    assert process.stdout
    assert process.stderr

    stdout_capture_task = asyncio.create_task(
        capture(process.stdout, STDOUT_FD, output)
    )

    stderr_capture_task = asyncio.create_task(
        capture(process.stderr, STDERR_FD, output)
    )

    async def capture_until_exit() -> int:
        await asyncio.wait({stdout_capture_task, stderr_capture_task})
        return await process.wait()

    try:
        exit_code = await asyncio.wait_for(capture_until_exit(), timeout)
    except (TimeoutError, asyncio.TimeoutError):  # noqa: UP041
        process.kill()

    formatted_output = "".join([chunk for (_, chunk) in output]).strip() + "\n\n"

    if exit_code is None:
        formatted_output += f"EXIT CODE: not available - Exponent stopped the command after {timeout} seconds"
    else:
        formatted_output += f"EXIT CODE: {exit_code}"

    return formatted_output


async def capture(
    stream: asyncio.StreamReader, fd: int, output: list[tuple[int, str]]
) -> None:
    while True:
        data = await stream.read(4096)

        if not data:
            break

        output.append((fd, data.decode()))



import logging
import os
import re
import subprocess
from collections.abc import Callable

from diff_match_patch import diff_match_patch
from pydantic import BaseModel

from exponent.core.remote_execution.types import (
    WRITE_STRATEGY_FULL_FILE_REWRITE,
    WRITE_STRATEGY_NATURAL_EDIT,
    WRITE_STRATEGY_SEARCH_REPLACE,
    WRITE_STRATEGY_UDIFF,
    FileWriteRequest,
    FileWriteResponse,
)
from exponent.core.remote_execution.utils import assert_unreachable

logger = logging.getLogger(__name__)


class FileEditResult(BaseModel):
    content: str | None
    failed_edits: list[tuple[str, str]]


def execute_file_write(
    request: FileWriteRequest, working_directory: str
) -> FileWriteResponse:
    write_strategy = request.write_strategy
    if write_strategy == WRITE_STRATEGY_FULL_FILE_REWRITE:
        result = execute_full_file_rewrite(
            request.file_path, request.content, working_directory
        )
    elif write_strategy == WRITE_STRATEGY_UDIFF:
        result = execute_udiff_edit(
            request.file_path, request.content, working_directory
        )
    elif write_strategy == WRITE_STRATEGY_SEARCH_REPLACE:
        result = execute_search_replace_edit(
            request.file_path, request.content, working_directory
        )
    elif write_strategy == WRITE_STRATEGY_NATURAL_EDIT:
        result = execute_full_file_rewrite(
            request.file_path, request.content, working_directory
        )
    else:
        assert_unreachable(write_strategy)
    return FileWriteResponse(
        content=result,
        correlation_id=request.correlation_id,
    )


def lint_file(file_path: str, working_directory: str) -> str:
    try:
        # Construct the absolute path
        full_file_path = os.path.join(working_directory, file_path)

        # Run ruff check --fix on the file
        result = subprocess.run(
            ["ruff", "check", "--fix", full_file_path],
            capture_output=True,
            text=True,
            check=True,
        )

        # If the subprocess ran successfully, return a success message
        return f"Lint results:\n\n{result.stdout}\n\n{result.stderr}"
    except Exception as e:  # noqa: BLE001
        # For any other errors, return a generic error message
        return f"An error occurred while linting: {e!s}"


def format_results(file_path: str, content: str, working_directory: str) -> str:
    footer = ""

    # In case you want linting
    # if is_editable_install():
    #     lint_result = lint_file(file_path, working_directory)
    #     footer = f"\n\n{lint_result}"

    return content + footer


def execute_full_file_rewrite(
    file_path: str, content: str, working_directory: str
) -> str:
    try:
        # Construct the absolute path
        full_file_path = os.path.join(working_directory, file_path)

        # Check if the directory exists, if not, create it
        os.makedirs(os.path.dirname(full_file_path), exist_ok=True)

        # Determine if the file exists and write the new content
        if os.path.exists(full_file_path):
            with open(full_file_path, "w") as file:
                file.write(content)
            result = f"Modified file {file_path}"
        else:
            with open(full_file_path, "w") as file:
                file.write(content)
            result = f"Created file {file_path}"

        return format_results(file_path, result, working_directory)

    except Exception as e:  # noqa: BLE001
        return f"An error occurred: {e!s}"


def execute_udiff_edit(file_path: str, content: str, working_directory: str) -> str:
    result = execute_partial_edit(file_path, content, working_directory, apply_udiff)
    return format_results(file_path, result, working_directory)


def execute_search_replace_edit(
    file_path: str, content: str, working_directory: str
) -> str:
    result = execute_partial_edit(
        file_path, content, working_directory, apply_all_search_replace
    )
    return format_results(file_path, result, working_directory)


def execute_partial_edit(
    file_path: str,
    edit_content: str,
    working_directory: str,
    edit_function: Callable[[str, str], FileEditResult],
) -> str:
    try:
        # Construct the absolute path
        full_file_path = os.path.join(working_directory, file_path)

        # Check if the directory exists, if not, create it
        os.makedirs(os.path.dirname(full_file_path), exist_ok=True)

        # Determine if the file exists and write the new content
        file_content, created = read_or_init_file(full_file_path)

        success = open_file_and_apply_edit(
            file_path=full_file_path,
            file_content=file_content,
            edit_content=edit_content,
            edit_function=edit_function,
        )

        if success:
            verb = "Created" if created else "Modified"
            return f"{verb} file {file_path}"
        else:
            verb = "create" if created else "modify"
            return f"Failed to {verb} file {file_path}"

    except Exception as e:
        raise e


def read_or_init_file(file_path: str) -> tuple[str, bool]:
    if not os.path.exists(file_path):
        with open(file_path, "w") as f:
            f.write("")
        return "", True

    with open(file_path) as f:
        return f.read(), False


def open_file_and_apply_edit(
    file_path: str,
    file_content: str,
    edit_content: str,
    edit_function: Callable[[str, str], FileEditResult],
) -> bool:
    result = edit_function(file_content, edit_content)

    if not result.content:
        return False

    with open(file_path, "w") as file:
        file.write(result.content)

    return True


def try_search_replace(existing_content: str, search: str, replace: str) -> str | None:
    # Try simple search and replace first
    new_content = simple_search_and_replace(existing_content, search, replace)
    if new_content:
        print("Applied simple search and replace")
        return new_content

    return None


def try_diff_patch(existing_content: str, search: str, replace: str) -> str | None:
    new_content = diff_patch_search_and_replace(existing_content, search, replace)
    if new_content:
        print("Applied diff patch search and replace")
        return new_content

    return None


def apply_udiff(existing_content: str, diff_content: str) -> FileEditResult:
    hunks = get_raw_udiff_hunks(diff_content)

    for hunk in hunks:
        if not hunk:
            continue

        search, replace = split_hunk_for_search_and_replace(hunk)

        # Exact match
        new_content = try_search_replace(existing_content, search, replace)
        if new_content is not None:
            print("Applied successfully!")
            return FileEditResult(content=new_content, failed_edits=[])

        # Fuzzy match
        new_content = try_diff_patch(existing_content, search, replace)
        if new_content is not None:
            print("Applied successfully!")
            return FileEditResult(content=new_content, failed_edits=[])

        print("Failed to apply hunk, exiting!")
        return FileEditResult(content=None, failed_edits=[(search, replace)])

    return FileEditResult(content=existing_content, failed_edits=[])


def get_raw_udiff_hunks(content: str) -> list[list[str]]:
    lines = content.splitlines(keepends=True)
    hunks: list[list[str]] = []
    current_hunk: list[str] = []
    for line in lines:
        if line.startswith("@@"):
            if current_hunk:
                hunks.append(current_hunk)
                current_hunk = []
        else:
            current_hunk.append(line)
    if current_hunk:
        hunks.append(current_hunk)
    return hunks


def split_hunk_for_search_and_replace(hunk: list[str]) -> tuple[str, str]:
    search_lines = []
    replace_lines = []

    search_prefixes = ["-", " "]
    replace_prefixes = ["+", " "]
    for line in hunk:
        if not line:
            continue
        prefix, content = line[0], line[1:]
        if not content:
            continue
        if prefix in search_prefixes:
            search_lines.append(content)
        if prefix in replace_prefixes:
            replace_lines.append(content)
    return "".join(search_lines), "".join(replace_lines)


def simple_search_and_replace(content: str, search: str, replace: str) -> str | None:
    if content.count(search) >= 1:
        return content.replace(search, replace)
    return None


def diff_patch_search_and_replace(
    content: str, search: str, replace: str
) -> str | None:
    patcher = diff_match_patch()
    # 3 second tieout for computing diffs
    patcher.Diff_Timeout = 3
    patcher.Match_Threshold = 0.95
    patcher.Match_Distance = 500
    patcher.Match_MaxBits = 128
    patcher.Patch_Margin = 32
    search_vs_replace_diff = patcher.diff_main(search, replace, False)

    # Simplify the diff as much as possible
    patcher.diff_cleanupEfficiency(search_vs_replace_diff)
    patcher.diff_cleanupSemantic(search_vs_replace_diff)

    original_vs_search_diff = patcher.diff_main(search, content)
    new_diffs = patcher.patch_make(search, search_vs_replace_diff)
    # Offset the search vs. replace diffs with the offset
    # of the search diff within the original content.
    for new_diff in new_diffs:
        new_diff.start1 = patcher.diff_xIndex(original_vs_search_diff, new_diff.start1)
        new_diff.start2 = patcher.diff_xIndex(original_vs_search_diff, new_diff.start2)

    new_content, successes = patcher.patch_apply(new_diffs, content)
    if not all(successes):
        return None

    return str(new_content)


SEARCH_REPLACE_RE = re.compile(
    r"[^<>]*<<<+\s*SEARCH\n((?P<search>.*?)\n)?===+\n((?P<replace>.*?)\n)?>>>+\s*?REPLACE\s*?[^<>]*",
    re.DOTALL,
)


def apply_search_replace(
    result: str, search: str | None, replace: str | None
) -> str | None:
    if search and replace:
        new_result = try_search_replace(result, search, replace)
        return new_result

    elif replace and not result:
        return replace

    return None


def apply_all_search_replace(existing_content: str, sr_content: str) -> FileEditResult:
    # Same as apply_search_replace, but applies all search and replace pairs
    # in the sr_content to the existing_content

    result = existing_content
    failed_edits: list[tuple[str, str]] = []

    for match in SEARCH_REPLACE_RE.finditer(sr_content):
        match_dict = match.groupdict()
        search, replace = match_dict.get("search"), match_dict.get("replace")

        new_result = apply_search_replace(result, search, replace)
        if new_result is None:
            failed_edits.append((search or "", replace or ""))
            continue

        result = new_result

    return FileEditResult(content=result, failed_edits=failed_edits)



CREATE_GITHUB_CONFIG_MUTATION: str = """
    mutation CreateGithubConfig(
        $githubPat: String!,
    ) {
        createGithubConfig(
            githubPat: $githubPat
        ) {
            __typename
            ... on GithubConfig {
                githubConfigUuid
                githubPat
            }
        }
    }
"""



AUTHENTICATED_USER_SUBSCRIPTION = """
    subscription {
            testAuthenticatedUser {
                __typename
                ... on UnauthenticatedError {
                    message
                }
                ...on Error {
                    message
                }
                ... on User {
                    userUuid
                }
            }
        }
"""

CHAT_EVENTS_SUBSCRIPTION = """
  subscription ChatEvents(
    $prompt: Prompt
    $codeBlockConfirmation: CodeBlockConfirmationResponse
    $fileWriteConfirmation: FileWriteConfirmationResponse
    $commandConfirmation: CommandConfirmationResponse
    $chatUuid: String!
    $parentUuid: String
    $model: LiteLLMModels!
    $strategyNameOverride: StrategyName
    $useToolsConfig: UseToolsMode!
    $requireConfirmation: Boolean
    $strategyNameOverride: StrategyName
  ) {
    authenticatedChat(
      chatInput: {
        prompt: $prompt
        codeBlockConfirmation: $codeBlockConfirmation
        fileWriteConfirmation: $fileWriteConfirmation
        commandConfirmation: $commandConfirmation
      }
      parentUuid: $parentUuid
      chatConfig: {
        chatUuid: $chatUuid
        model: $model
        useToolsConfig: $useToolsConfig
        requireConfirmation: $requireConfirmation
        strategyNameOverride: $strategyNameOverride
      }
    ) {
      __typename
      ... on UnauthenticatedError {
        __typename
        message
      }
      ... on Error {
        __typename
        message
      }
      ... on MessageChunkEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        role
        content
      }
      ... on MessageEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        role
        content
        attachments {
          ... on FileAttachment {
            file {
              filePath
              workingDirectory
            }
            content
          }
          ... on URLAttachment {
            url
            content
          }
        }
      }
      ... on CodeBlockChunkEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        language
        content
      }
      ... on CodeBlockEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        language
        content
        requireConfirmation
      }
      ... on CodeBlockConfirmationEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        codeBlockUuid
        accepted
      }
      ... on CodeExecutionEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        codeBlockUuid
        content
      }
      ... on CodeExecutionStartEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        codeBlockUuid
      }
      ... on FileWriteChunkEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        filePath
        language
        writeStrategy
        content
        writeContent {
          ... on EditContent {
            content
          }
          ... on NaturalEditContent {
            newFile
            originalFile
            naturalEdit
            errorContent
          }
        }
      }
      ... on FileWriteEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        filePath
        language
        writeStrategy
        content
        writeContent {
          ... on EditContent {
            content
          }
          ... on NaturalEditContent {
            newFile
            originalFile
            naturalEdit
            errorContent
          }
        }
        requireConfirmation
      }
      ... on FileWriteConfirmationEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        fileWriteUuid
        accepted
      }
      ... on FileWriteResultEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        fileWriteUuid
        content
      }
      ... on FileWriteStartEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        fileWriteUuid
      }
      ... on CommandChunkEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        data {
          __typename
          ... on ThinkingCommandData {
            type
            content
          }
          ... on FileOpenCommandData {
            type
            filePath
            language
          }
          ... on FileReadCommandData {
            type
            filePath
            language
          }
        }
      }
      ... on CommandEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        data {
          __typename
          ... on ThinkingCommandData {
            type
            content
          }
          ... on FileReadCommandData {
            type
            filePath
            language
          }
        }
        requireConfirmation
      }
      ... on CommandConfirmationEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        commandUuid
        accepted
      }
      ... on CommandStartEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        commandUuid
      }
      ... on CommandResultEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
        commandUuid
        content
      }
      ... on RemoteStartEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
      }
      ... on RemoteEndEvent {
        __typename
        chatUuid
        eventUuid
        parentUuid
      }
    }
  }
"""



EVENTS_FOR_CHAT_QUERY: str = """query EventsForChat($chatUuid: String!) {
  eventsForChat(chatUuid: $chatUuid) {
    ... on EventHistory {
      events {
        ... on GraphExponentEvent {
          eventUuid
        }
      }
    }
  }
}
"""



import json
import os
import sys
import time
from typing import Any, cast

import click
from pydantic import ValidationError

from exponent.commands.common import create_chat, redirect_to_login, run_until_complete
from exponent.commands.settings import use_settings
from exponent.commands.types import exponent_cli_group
from exponent.commands.utils import (
    launch_exponent_browser,
    print_exponent_message,
    write_template_exponent_cloud_config,
)
from exponent.core.config import Environment, ExponentCloudConfig, Settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.get_cloud_chat_state_query import GET_CLOUD_CHAT_STATE_QUERY
from exponent.core.graphql.mutations import MOVE_CHAT_TO_CLOUD_MUTATION
from exponent.core.remote_execution.types import DevboxConnectedState
from exponent.core.runloop import RunloopClient


@exponent_cli_group()
def cloud_cli() -> None:
    pass


@cloud_cli.command()
@use_settings
def cloud(
    settings: Settings,
) -> None:
    """Start an Exponent session in the cloud."""
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        start_cloud(
            settings.environment,
            settings.api_key,
            settings.base_url,
            settings.base_api_url,
        )
    )


async def start_cloud(
    environment: Environment,
    api_key: str,
    base_url: str,
    base_api_url: str,
) -> None:
    chat_uuid = await create_chat(api_key, base_api_url)

    if chat_uuid is None:
        return

    graphql_client = GraphQLClient(api_key=api_key, base_api_url=base_api_url)
    response = await graphql_client.query(
        MOVE_CHAT_TO_CLOUD_MUTATION, {"chatUuid": chat_uuid}, "MoveChatToCloud"
    )

    if response["moveChatToCloud"]["chatUuid"] != chat_uuid:
        click.secho(
            f"Failed to initalize cloud chat: {response['moveChatToCloud']}",
            fg="red",
            bold=True,
        )
        return

    click.secho(
        "Chat created. Waiting for cloud container to spin up...", fg="green", bold=True
    )

    while True:
        chatState = await fetch_cloud_chat_state(
            api_key,
            base_api_url,
            chat_uuid,
        )
        connectedState = chatState["cloudChatState"]["connectedState"]
        if connectedState == DevboxConnectedState.CONNECTED.value:
            break

        time.sleep(1)

    print_exponent_message(base_url, chat_uuid)
    launch_exponent_browser(environment, base_url, chat_uuid)


@cloud_cli.command(hidden=True)
@click.option(
    "--chat-uuid",
    help="Chat UUID to get state for.",
)
@use_settings
def cloud_chat_state(
    settings: Settings,
    chat_uuid: str,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        print_cloud_chat_state(
            settings.api_key,
            settings.base_api_url,
            chat_uuid,
        )
    )


async def print_cloud_chat_state(
    api_key: str,
    base_api_url: str,
    chat_uuid: str,
) -> None:
    response = await fetch_cloud_chat_state(
        api_key,
        base_api_url,
        chat_uuid,
    )
    click.secho(
        json.dumps(response, indent=2),
        fg="yellow",
    )


async def fetch_cloud_chat_state(
    api_key: str,
    base_api_url: str,
    chat_uuid: str,
) -> dict[str, Any]:
    graphql_client = GraphQLClient(api_key=api_key, base_api_url=base_api_url)
    return cast(
        dict[str, Any],
        await graphql_client.query(
            GET_CLOUD_CHAT_STATE_QUERY, {"chatUuid": chat_uuid}, "GetCloudChatState"
        ),
    )


@cloud_cli.command(hidden=True)
@click.option(
    "--prompt",
    help="Start a chat with a given prompt.",
)
def deprecated_cloud(
    settings: Settings,
    prompt: str | None = None,
) -> None:
    if not settings.api_key:
        redirect_to_login(settings)
        return

    run_until_complete(
        deprecated_start_cloud(
            settings.environment,
            settings.api_key,
            settings.base_url,
            settings.base_api_url,
            prompt=prompt,
        )
    )


async def deprecated_start_cloud(
    environment: Environment,
    api_key: str,
    base_url: str,
    base_api_url: str,
    prompt: str | None = None,
) -> None:
    current_working_directory = os.getcwd()

    # Check if an `.exponent.cloud.json` file exists.
    # If so, use it. If not, write a template file and exit.
    file_path = os.path.join(current_working_directory, ".exponent.cloud.json")
    if not os.path.exists(file_path):
        click.secho("No `.exponent.cloud.json` file found, creating one to fill out...")
        write_template_exponent_cloud_config(file_path)
        click.secho(
            "`.exponent.cloud.json` file created, fill out the required fields "
            "and run this command again."
        )
        return

    with open(file_path) as f:
        try:
            exponent_cloud_config = ExponentCloudConfig.model_validate_json(f.read())
        except ValidationError as e:
            click.secho(f"Error in parsing `.exponent.cloud.json`: {e}", fg="red")
            return

    chat_uuid = await create_chat(api_key, base_api_url)

    if chat_uuid is None:
        return

    click.secho(
        "Chat created. Waiting for cloud container to spin up...", fg="green", bold=True
    )

    runloop_client = RunloopClient(
        api_key=exponent_cloud_config.runloop_api_key,
    )

    def join_commands(commands: list[str]) -> str:
        return " && ".join(commands)

    if prompt:
        exponent_command = (
            f'exponent run --prod --chat-id {chat_uuid} --prompt \\"{prompt}\\"'
        )
    else:
        exponent_command = f"exponent run --prod --chat-id {chat_uuid}"

    devbox = await runloop_client.create_devbox(
        entrypoint="/home/user/run.sh",
        environment_variables={"GH_TOKEN": exponent_cloud_config.gh_token},
        setup_commands=[
            join_commands(exponent_cloud_config.repo_specific_setup_commands),
            f'echo "cd /home/user/{exponent_cloud_config.repo_name} && source .venv/bin/activate && uv pip install exponent-run && exponent login --prod --key {api_key} && {exponent_command}" > /home/user/run.sh',
            "chmod +x /home/user/run.sh",
        ],
    )

    # Step 3. Poll Runloop for container spinup and log status
    while True:
        current_devbox = await runloop_client.get_devbox(devbox["id"])
        if current_devbox["status"] != "running":
            print(
                f"Container {devbox['id']} is loading, waiting.... Current status is {current_devbox['status']}"
            )
        elif current_devbox["status"] == "failure":
            click.secho("Devbox failed to start", fg="red", bold=True)
            sys.exit(1)
        else:
            break

        time.sleep(1)

    # Step 4. Open the chat in the browser
    print_exponent_message(base_url, chat_uuid)
    launch_exponent_browser(environment, base_url, chat_uuid)

    # Step 5. Wait for user input with the message "Stop Runloop?"
    input("Stop cloud container?: [press enter to continue]")

    # Step 6. Stop Runloop
    await runloop_client.shutdown_devbox(devbox["id"])
    click.secho("Cloud container stopped", fg="green", bold=True)


@cloud_cli.command(hidden=True)
@click.option(
    "--devbox-id",
    type=str,
    help="Devbox ID to get logs from.",
    required=True,
    prompt=True,
)
def cloud_logs(
    devbox_id: str,
    settings: Settings,
) -> None:
    run_until_complete(start_cloud_logs(devbox_id))


async def start_cloud_logs(
    devbox_id: str,
) -> None:
    current_working_directory = os.getcwd()

    # Check if an `.exponent.cloud.json` file exists.
    # If so, use it. If not, write a template file and exit.
    file_path = os.path.join(current_working_directory, ".exponent.cloud.json")
    if not os.path.exists(file_path):
        click.secho("No `.exponent.cloud.json` file found, creating one to fill out...")
        write_template_exponent_cloud_config(file_path)
        click.secho(
            "`.exponent.cloud.json` file created, fill out the required fields "
            "and run this command again."
        )
        return

    with open(file_path) as f:
        try:
            exponent_cloud_config = ExponentCloudConfig.model_validate_json(f.read())
        except ValidationError as e:
            click.secho(f"Error in parsing `.exponent.cloud.json`: {e}", fg="red")
            return

    runloop_client = RunloopClient(
        api_key=exponent_cloud_config.runloop_api_key,
    )

    response = await runloop_client.devbox_logs(devbox_id)
    for log in response["logs"]:
        click.secho(log["message"])



import json
import logging
from collections.abc import Awaitable, Callable
from functools import wraps
from typing import (
    NoReturn,
    ParamSpec,
    TypeGuard,
    TypeVar,
    cast,
    overload,
)

from exponent.version import get_version
import websockets
import websockets.exceptions
from httpx import Response
from pydantic import BaseModel
from sentry_sdk.serializer import serialize
from sentry_sdk.utils import (
    event_from_exception,
    exc_info_from_error,
)

from exponent.core.remote_execution.types import (
    SUPPORTED_LANGUAGES,
    CLIErrorLog,
    CodeExecutionRequest,
    CodeExecutionResponse,
    CommandRequest,
    CommandResponse,
    FileWriteRequest,
    FileWriteResponse,
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GetFileAttachmentRequest,
    GetFileAttachmentResponse,
    GetFileAttachmentsRequest,
    GetFileAttachmentsResponse,
    GetMatchingFilesRequest,
    GetMatchingFilesResponse,
    ListFilesRequest,
    ListFilesResponse,
    RemoteExecutionMessage,
    RemoteExecutionMessageData,
    RemoteExecutionRequest,
    RemoteExecutionRequestData,
    RemoteExecutionRequestType,
    RemoteExecutionResponse,
    RemoteExecutionResponseData,
    RemoteExecutionResponseType,
    SupportedLanguage,
    SystemContextRequest,
    SystemContextResponse,
)

### Serde


def deserialize_message_data(
    message_data: RemoteExecutionMessageData | str,
) -> RemoteExecutionMessage:
    if isinstance(message_data, str):
        message_data = RemoteExecutionMessageData.model_validate_json(message_data)
    if message_data.direction == "request":
        return deserialize_request_data(cast(RemoteExecutionRequestData, message_data))
    elif message_data.direction == "response":
        return deserialize_response_data(
            cast(RemoteExecutionResponseData, message_data)
        )
    else:
        # type checking trick, if you miss a namespace then
        # this won't typecheck due to the input parameter
        # having a potential type other than no-return
        assert_unreachable(message_data.direction)


def deserialize_request_data(
    request_data: RemoteExecutionRequestData | str,
) -> RemoteExecutionRequestType:
    request: RemoteExecutionRequestType
    if isinstance(request_data, str):
        request_data = RemoteExecutionRequestData.model_validate_json(request_data)
    if request_data.direction != "request":
        raise ValueError(f"Expected request, but got {request_data.direction}")
    if request_data.namespace == "code_execution":
        request = CodeExecutionRequest.model_validate_json(request_data.message_data)
    elif request_data.namespace == "file_write":
        request = FileWriteRequest.model_validate_json(request_data.message_data)
    elif request_data.namespace == "list_files":
        request = ListFilesRequest.model_validate_json(request_data.message_data)
    elif request_data.namespace == "get_file_attachment":
        request = GetFileAttachmentRequest.model_validate_json(
            request_data.message_data
        )
    elif request_data.namespace == "get_file_attachments":
        request = GetFileAttachmentsRequest.model_validate_json(
            request_data.message_data
        )
    elif request_data.namespace == "get_matching_files":
        request = GetMatchingFilesRequest.model_validate_json(request_data.message_data)
    elif request_data.namespace == "system_context":
        request = SystemContextRequest.model_validate_json(request_data.message_data)
    elif request_data.namespace == "get_all_tracked_files":
        request = GetAllTrackedFilesRequest.model_validate_json(
            request_data.message_data
        )
    elif request_data.namespace == "command":
        request = CommandRequest.model_validate_json(request_data.message_data)
    else:
        # type checking trick, if you miss a namespace then
        # this won't typecheck due to the input parameter
        # having a potential type other than no-return
        request = assert_unreachable(request_data.namespace)
    return truncate_message(request)


def deserialize_response_data(
    response_data: RemoteExecutionResponseData | str,
) -> RemoteExecutionResponseType:
    response: RemoteExecutionResponseType
    if isinstance(response_data, str):
        response_data = RemoteExecutionResponseData.model_validate_json(response_data)
    if response_data.direction != "response":
        raise ValueError(f"Expected response, but got {response_data.direction}")
    if response_data.namespace == "code_execution":
        response = CodeExecutionResponse.model_validate_json(response_data.message_data)
    elif response_data.namespace == "file_write":
        response = FileWriteResponse.model_validate_json(response_data.message_data)
    elif response_data.namespace == "list_files":
        response = ListFilesResponse.model_validate_json(response_data.message_data)
    elif response_data.namespace == "get_matching_files":
        response = GetMatchingFilesResponse.model_validate_json(
            response_data.message_data
        )
    elif response_data.namespace == "get_file_attachment":
        response = GetFileAttachmentResponse.model_validate_json(
            response_data.message_data
        )
    elif response_data.namespace == "get_file_attachments":
        response = GetFileAttachmentsResponse.model_validate_json(
            response_data.message_data
        )
    elif response_data.namespace == "system_context":
        response = SystemContextResponse.model_validate_json(response_data.message_data)
    elif response_data.namespace == "get_all_tracked_files":
        response = GetAllTrackedFilesResponse.model_validate_json(
            response_data.message_data
        )
    elif response_data.namespace == "command":
        response = CommandResponse.model_validate_json(response_data.message_data)
    else:
        # type checking trick, if you miss a namespace then
        # this won't typecheck due to the input parameter
        # having a potential type other than no-return
        response = assert_unreachable(response_data.namespace)
    return truncate_message(response)


def serialize_message(response: RemoteExecutionMessage) -> str:
    truncated_response = truncate_message(response)
    message = RemoteExecutionMessageData(
        namespace=response.namespace,
        direction=response.direction,
        message_data=truncated_response.model_dump_json(),
    )
    return message.model_dump_json()


### API Serde


TModel = TypeVar("TModel", bound=BaseModel)


async def deserialize_list_api_response(
    response_data: Response,
) -> list[RemoteExecutionRequestType]:
    response_json = response_data.json()
    results = []
    for result in response_json:
        results.append(deserialize_request_data(result))
    return results


async def deserialize_api_response(
    response: Response,
    data_model: type[TModel],
) -> TModel:
    response_json = response.json()
    return data_model.model_validate(response_json)


### Validation


ResponseT = TypeVar("ResponseT", bound=RemoteExecutionResponse)


def response_type_is_valid_for_request(
    response: RemoteExecutionResponseType, request: RemoteExecutionRequest[ResponseT]
) -> TypeGuard[ResponseT]:
    if request.namespace != response.namespace or response.direction != "response":
        raise ValueError(
            f"Expected {request.namespace}.response, but got {response.namespace}.{response.direction}"
        )
    return True


def assert_unreachable(x: NoReturn) -> NoReturn:
    assert False, f"Unhandled type: {type(x).__name__}"


def is_supported_language(language: str) -> TypeGuard[SupportedLanguage]:
    if language not in SUPPORTED_LANGUAGES:
        return False
    return True


### Truncation


OUTPUT_CHARACTER_MAX = 90_000  # A tad over ~8k tokens
TRUNCATION_MESSAGE_CHARS = (
    "(Output truncated, only showing the first {remaining_chars} characters)"
)
TRUNCATION_MESSAGE_LINES = (
    "(Output truncated, only showing the first {remaining_lines} lines)"
)
LONGEST_TRUNCATION_MESSAGE_LEN = (
    len(TRUNCATION_MESSAGE_CHARS.format(remaining_chars=OUTPUT_CHARACTER_MAX)) + 1
)


def truncate_output(output: str, character_limit: int = OUTPUT_CHARACTER_MAX) -> str:
    output_length = len(output)
    # When under the character limit, return the output as is.
    # Note we're adding the length of the truncation message + 1
    # to the character limit to account for the fact that the
    # truncation message will be added to the output + a newline.
    # In case we want to run truncation logic both client side
    # and server side, we want to account for the truncation
    # message length to avoid weird double truncation overlap.
    if output_length <= character_limit + LONGEST_TRUNCATION_MESSAGE_LEN:
        return output

    # Attempt to trim whole lines until we're under
    # the character limit.
    lines = output.split("\n")
    while output_length > character_limit:
        last_line = lines.pop()
        # +1 to account for the newline
        output_length -= len(last_line) + 1

    if not lines:
        # If we truncated all the lines, then we have
        # have some ridiculous long line at the start
        # of the output so we'll just truncate by
        # character count to retain something.
        output = output[:character_limit]
        # Format the truncation message accordingly
        truncation_message = TRUNCATION_MESSAGE_CHARS.format(
            remaining_chars=character_limit,
        )
    else:
        # Otherwise, just join the lines back together
        output = "\n".join(lines)
        truncation_message = TRUNCATION_MESSAGE_LINES.format(
            remaining_lines=len(lines),
        )

    return f"{output}\n{truncation_message}"


@overload
def truncate_message(response: CodeExecutionRequest) -> CodeExecutionRequest: ...
@overload
def truncate_message(response: CodeExecutionResponse) -> CodeExecutionResponse: ...
@overload
def truncate_message(response: FileWriteRequest) -> FileWriteRequest: ...
@overload
def truncate_message(response: FileWriteResponse) -> FileWriteResponse: ...
@overload
def truncate_message(
    response: GetFileAttachmentRequest,
) -> GetFileAttachmentRequest: ...
@overload
def truncate_message(
    response: GetFileAttachmentResponse,
) -> GetFileAttachmentResponse: ...
@overload
def truncate_message(response: ListFilesRequest) -> ListFilesRequest: ...
@overload
def truncate_message(response: ListFilesResponse) -> ListFilesResponse: ...
@overload
def truncate_message(response: GetMatchingFilesRequest) -> GetMatchingFilesRequest: ...
@overload
def truncate_message(
    response: GetMatchingFilesResponse,
) -> GetMatchingFilesResponse: ...
@overload
def truncate_message(response: SystemContextRequest) -> SystemContextRequest: ...
@overload
def truncate_message(response: SystemContextResponse) -> SystemContextResponse: ...
@overload
def truncate_message(
    response: RemoteExecutionRequestType,
) -> RemoteExecutionRequestType: ...
@overload
def truncate_message(
    response: RemoteExecutionResponseType,
) -> RemoteExecutionResponseType: ...
@overload
def truncate_message(response: RemoteExecutionMessage) -> RemoteExecutionMessage: ...


def truncate_message(
    response: RemoteExecutionMessage,
) -> RemoteExecutionMessage:
    if isinstance(response, CodeExecutionResponse | GetFileAttachmentResponse):
        response.content = truncate_output(response.content)
    elif isinstance(response, GetFileAttachmentsResponse):
        for file_attachment in response.file_attachments:
            file_attachment.content = truncate_output(file_attachment.content)
    return response


### Error Handling


def format_attachment_data(attachment_lines: list[str] | None = None) -> str | None:
    if not attachment_lines:
        return None
    log_attachment_str = "\n".join(attachment_lines)
    return log_attachment_str


def format_error_log(
    exc: Exception,
    chat_uuid: str | None = None,
    attachment_lines: list[str] | None = None,
) -> CLIErrorLog | None:
    exc_info = exc_info_from_error(exc)
    event, _ = event_from_exception(exc_info)
    attachment_data = format_attachment_data(attachment_lines)
    version = get_version()

    try:
        event_data = json.dumps(serialize(event))  # type: ignore
    except json.JSONDecodeError:
        return None

    return CLIErrorLog(
        event_data=event_data,
        attachment_data=attachment_data,
        version=version,
        chat_uuid=chat_uuid,
    )


### Websockets


ws_logger = logging.getLogger("WebsocketUtils")


TParam = ParamSpec("TParam")


def ws_retry(
    connection_name: str,
    max_retries: int = 5,
) -> Callable[[Callable[TParam, Awaitable[None]]], Callable[TParam, Awaitable[None]]]:
    connection_name = connection_name.capitalize()
    reconnect_msg = f"{connection_name} reconnecting."
    disconnect_msg = f"{connection_name} connection closed."
    max_disconnect_msg = (
        f"{connection_name} connection closed {max_retries} times, exiting."
    )

    def decorator(
        f: Callable[TParam, Awaitable[None]],
    ) -> Callable[TParam, Awaitable[None]]:
        @wraps(f)
        async def wrapped(*args: TParam.args, **kwargs: TParam.kwargs) -> None:
            i = 0

            while True:
                try:
                    return await f(*args, **kwargs)
                except (websockets.exceptions.ConnectionClosed, TimeoutError) as e:
                    # Warn on disconnect
                    ws_logger.warning(disconnect_msg)

                    if i >= max_retries:
                        # We've reached the max number of retries,
                        # log an error and reraise
                        ws_logger.warning(max_disconnect_msg)
                        raise e

                    # Increment the retry count
                    i += 1
                    # Notify the user that we're reconnecting
                    ws_logger.warning(reconnect_msg)
                    continue

        return wrapped

    return decorator



import asyncio
import logging
import queue
import re
import subprocess
import threading
import time
from typing import Any

from jupyter_client.client import KernelClient
from jupyter_client.manager import KernelManager

from exponent.core.remote_execution.types import PythonEnvInfo

logger = logging.getLogger(__name__)


class IOChannelHandler:
    ESCAPE_SEQUENCE = re.compile(r"\x1B\[[0-?]*[ -/]*[@-~]")

    def __init__(self) -> None:
        self.output_buffer: queue.Queue[str] = queue.Queue()

    def add_message(self, message: dict[str, Any]) -> None:
        logger.debug(f"Jupyter kernel message received: {message}")
        output = None
        if message["msg_type"] == "stream":
            output = message["content"]["text"]
        elif message["msg_type"] == "error":
            raw_content = "\n".join(message["content"]["traceback"])
            content = self.ESCAPE_SEQUENCE.sub("", raw_content)
            output = content
        if output:
            print(output)
            self.output_buffer.put(output)

    @staticmethod
    def is_idle(message: dict[str, Any]) -> bool:
        return bool(
            message["header"]["msg_type"] == "status"
            and message["content"]["execution_state"] == "idle"
        )


class Kernel:
    def __init__(self, working_directory: str) -> None:
        self._manager: KernelManager | None = None
        self._client: KernelClient | None = None
        self.io_handler: IOChannelHandler = IOChannelHandler()
        self.should_halt = False
        self.working_directory = working_directory

    @property
    def manager(self) -> KernelManager:
        if not self._manager:
            self._manager = KernelManager(kernel_name="python3")
            self._manager.start_kernel(cwd=self.working_directory)
        return self._manager

    @property
    def client(self) -> KernelClient:
        if not self._client:
            self._client = self.manager.client()

            while not self._client.is_alive():
                time.sleep(0.1)

            self._client.start_channels()
        return self._client

    async def wait_for_ready(self, timeout: int = 5) -> None:
        manager = self.manager
        start_time = time.time()
        while not manager.is_alive():
            if time.time() - start_time > timeout:
                raise Exception("Kernel took too long to start")
            await asyncio.sleep(0.05)
        await asyncio.sleep(0.5)

    def iopub_listener(self, client: KernelClient) -> None:
        while True:
            if self.should_halt:
                logger.info("Received halt signal, stoping IO listener thread.")
                break
            try:
                msg = client.iopub_channel.get_msg(timeout=1)
                logger.debug(f"Received message from kernel: {msg}")
                self.io_handler.add_message(msg)
                if self.io_handler.is_idle(msg):
                    logger.debug("Kernel is idle, setting halt signal.")
                    self.should_halt = True
                    break
            except queue.Empty:
                continue
            except Exception as e:  # noqa: BLE001 - TODO: Deep audit potential exceptions
                logger.info(
                    f"Error getting message from kernel, halting io thread: {e}"
                )
                self.should_halt = True

    async def execute_code(self, code: str) -> str:
        # Make sure to clear this flag just in case
        self.should_halt = False
        # Grab the client and make sure it's ready
        await self.wait_for_ready()
        client = self.client
        client.connect_iopub()
        # Start separate thread for buffering kernel output
        iopub_thread = threading.Thread(target=self.iopub_listener, args=(client,))
        logger.info("Starting IO listener thread.")
        iopub_thread.start()

        logger.info("Executing code in kernel.")
        client.execute(code)

        results = []
        while True:
            if not self.io_handler.output_buffer.empty():
                output = self.io_handler.output_buffer.get()
                logger.info("Execution output: %s", output)
                results.append(output)
            if self.should_halt:
                break
            await asyncio.sleep(0.05)

        return "\n".join(results)

    def close(self) -> None:
        if self._client:
            self._client.stop_channels()
            self._client = None
        if self._manager:
            self._manager.shutdown_kernel()
            self._manager = None


async def execute_python(code: str, kernel: Kernel) -> str:
    return await kernel.execute_code(code)


### Environment Helpers


def get_python_env_info() -> PythonEnvInfo:
    return PythonEnvInfo(
        interpreter_path=get_active_python_interpreter_path(),
        interpreter_version=get_active_python_interpreter_version(),
    )


def get_active_python_interpreter_path() -> str | None:
    # Run the command `which python`
    result = subprocess.run(
        ["which", "python"],
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        return None
    return result.stdout.strip()


def get_active_python_interpreter_version() -> str | None:
    try:
        # Run the command `python -V`
        result = subprocess.run(
            ["python", "-V"], capture_output=True, text=True, check=False
        )
    except FileNotFoundError:
        # Python is not installed?
        return None
    # Capture the standard output and standard error
    stdout = result.stdout.strip()
    match = re.search(r"Python (\d+\.\d+\.\d+)", stdout)
    if match:
        return match.group(1)
    return None



import os
from collections.abc import Callable
from pathlib import Path
from typing import cast

import pygit2
from pygit2 import Repository, Tree

from exponent.core.remote_execution.files import file_walk, normalize_files
from exponent.core.remote_execution.types import (
    GetAllTrackedFilesRequest,
    GetAllTrackedFilesResponse,
    GitInfo,
    RemoteFile,
)

GIT_OBJ_COMMIT = 1


async def get_all_tracked_files(
    request: GetAllTrackedFilesRequest,
    working_directory: str,
) -> GetAllTrackedFilesResponse:
    return GetAllTrackedFilesResponse(
        correlation_id=request.correlation_id,
        files=await get_all_non_ignored_files(working_directory),
    )


async def get_all_non_ignored_files(working_directory: str) -> list[RemoteFile]:
    ignore_extra = None

    if get_repo(working_directory) is None:
        # If we have no git repo then use a default
        # list of ignore patterns to avoid returning
        # a million files
        ignore_extra = DEFAULT_IGNORES

    file_paths = file_walk(working_directory, ignore_extra=ignore_extra)

    return normalize_files(working_directory, await file_paths)


def get_repo(working_directory: str) -> Repository | None:
    try:
        return pygit2.Repository(working_directory)
    except pygit2.GitError:
        return None


def get_all_tracked_git_file_paths(
    repo: Repository, working_directory: str
) -> list[RemoteFile]:
    repo = pygit2.Repository(working_directory)
    files = get_tracked_files_in_dir(repo, working_directory)
    return normalize_files(working_directory, files)


def get_git_info(working_directory: str) -> GitInfo | None:
    try:
        repo = pygit2.Repository(working_directory)
    except pygit2.GitError:
        return None

    return GitInfo(
        branch=_get_git_branch(repo) or "<unknown branch>",
        remote=_get_git_remote(repo),
    )


def get_tracked_files_in_dir(
    repo: Repository, dir: str | Path, filter_func: Callable[[str], bool] | None = None
) -> list[str]:
    rel_path = get_path_relative_to_repo_root(repo, dir)
    dir_tree = get_git_subtree_for_dir(repo, dir)
    entries: list[str] = []
    if not dir_tree:
        return entries
    for entry in dir_tree:
        if not entry.name:
            continue
        entry_path = str(Path(f"{repo.workdir}/{rel_path}/{entry.name}"))
        if entry.type_str == "tree":
            entries.extend(get_tracked_files_in_dir(repo, entry_path, filter_func))
        elif entry.type_str == "blob":
            if not filter_func or filter_func(entry.name):
                entries.append(entry_path)
    return entries


def get_git_subtree_for_dir(repo: Repository, dir: str | Path) -> Tree | None:
    rel_path = get_path_relative_to_repo_root(repo, dir)

    try:
        head_commit = repo.head.peel(GIT_OBJ_COMMIT)
    except pygit2.GitError:
        # If the repo is empty, then the head commit will not exist
        return None
    head_tree: Tree = head_commit.tree

    if rel_path == Path("."):
        # If the relative path is the root of the repo, then
        # the head_tree is what we want. Note we do this because
        # Passing "." or "" as the path into the tree will raise.
        return head_tree
    return cast(Tree, head_tree[str(rel_path)])


def get_path_relative_to_repo_root(repo: Repository, path: str | Path) -> Path:
    path = Path(path).resolve()
    return path.relative_to(Path(repo.workdir).resolve())


def get_local_commit_hash() -> str:
    try:
        # Open the repository (assumes the current working directory is within the git repo)
        repo = pygit2.Repository(os.getcwd())

        # Get the current HEAD commit
        head = repo.head

        # Get the commit object and return its hash as a string
        return str(repo[head.target].id)
    except pygit2.GitError:
        return "unknown-local-commit"


def _get_git_remote(repo: pygit2.Repository) -> str | None:
    if repo.remotes:
        return str(repo.remotes[0].url)
    return None


def _get_git_branch(repo: pygit2.Repository) -> str | None:
    try:
        # Look for HEAD file in the .git directory
        head_path = os.path.join(repo.path, "HEAD")
        with open(head_path) as head_file:
            # The HEAD file content usually looks like: 'ref: refs/heads/branch_name'
            head_content = head_file.read().strip()
            if head_content.startswith("ref:"):
                return head_content.split("refs/heads/")[-1]
            else:
                return None
    except Exception:  # noqa: BLE001
        return None


DEFAULT_IGNORES = [
    ".venv",
    ".mypy_cache",
    ".pytest_cache",
    "node_modules",
    "venv",
    ".pyenv",
    "__pycache__",
    ".ipynb_checkpoints",
    ".vercel",
    "__pycache__/",
    "*.py[cod]",
    "*$py.class",
    ".env",
    "*.so",
    ".Python",
    "build/",
    "develop-eggs/",
    "dist/",
    "downloads/",
    "eggs/",
    ".eggs/",
    "lib/",
    "lib64/",
    "parts/",
    "sdist/",
    "var/",
    "wheels/",
    "pip-wheel-metadata/",
    "share/python-wheels/",
    "*.egg-info/",
    ".installed.cfg",
    "*.egg",
    "MANIFEST",
]



import asyncio
import difflib
import sys
from collections.abc import Callable, Coroutine, Iterable
from concurrent.futures import Future
from typing import Any

import click
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.key_binding import KeyBindings
from rich.console import Console
from rich.syntax import Syntax

from exponent.commands.common import (
    create_chat,
    inside_ssh_session,
    redirect_to_login,
    start_client,
)
from exponent.commands.settings import use_settings
from exponent.commands.types import (
    StrategyChoice,
    StrategyOption,
    exponent_cli_group,
)
from exponent.commands.utils import (
    ConnectionTracker,
    Spinner,
    start_background_event_loop,
)
from exponent.core.config import Settings
from exponent.core.graphql.client import GraphQLClient
from exponent.core.graphql.mutations import HALT_CHAT_STREAM_MUTATION
from exponent.core.graphql.queries import EVENTS_FOR_CHAT_QUERY
from exponent.core.graphql.subscriptions import CHAT_EVENTS_SUBSCRIPTION
from exponent.core.remote_execution.exceptions import ExponentError
from exponent.core.types.generated.strategy_info import (
    ENABLED_STRATEGY_INFO_LIST,
)

SLASH_COMMANDS = {
    "/help": "Show available commands",
    "/autorun": "Toggle autorun mode",
    "/web": "Move chat to a web browser",
}


@exponent_cli_group()
def shell_cli() -> None:
    pass


@shell_cli.command()
@click.option(
    "--model",
    help="LLM model",
    required=True,
    default="CLAUDE_3_POINT_5_SONNET",
)
@click.option(
    "--strategy",
    prompt=True,
    prompt_required=False,
    type=StrategyChoice(ENABLED_STRATEGY_INFO_LIST),
    cls=StrategyOption,
)
@click.option(
    "--autorun",
    is_flag=True,
    help="Enable autorun mode",
)
@click.option(
    "--chat-id",
    help="ID of an existing chat session to reconnect",
    required=False,
)
@click.option(
    "--prompt",
    help="Initial prompt",
)
@click.option(
    "--headless",
    is_flag=True,
    help="Run single prompt in headless mode",
)
@use_settings
def shell(
    settings: Settings,
    model: str,
    strategy: str,
    chat_id: str | None = None,
    autorun: bool = False,
    prompt: str | None = None,
    headless: bool = False,
) -> None:
    """Start an Exponent session in your current shell."""

    if not headless and not sys.stdin.isatty():
        print("Terminal not available, running in headless mode")
        headless = True

    if headless and not prompt:
        print("Error: --prompt option is required with headless mode")
        sys.exit(1)

    if not settings.api_key:
        redirect_to_login(settings)
        return

    api_key = settings.api_key
    base_api_url = settings.base_api_url
    gql_client = GraphQLClient(api_key, base_api_url)
    loop = start_background_event_loop()
    parent_event_uuid: str | None = None

    if chat_id is None:
        chat_uuid = asyncio.run_coroutine_threadsafe(
            create_chat(api_key, base_api_url), loop
        ).result()
    else:
        chat_uuid = chat_id

        parent_event_uuid = asyncio.run_coroutine_threadsafe(
            get_parent_event_uuid(gql_client, chat_uuid), loop
        ).result()

    if chat_uuid is None:
        sys.exit(1)

    connection_tracker = ConnectionTracker()

    client_coro = start_client(
        api_key, base_api_url, chat_uuid, connection_tracker=connection_tracker
    )

    if headless:
        assert prompt is not None

        chat = Chat(
            chat_uuid,
            parent_event_uuid,
            settings.base_url,
            gql_client,
            model,
            strategy,
            autorun,
            StaticView(),
        )

        client_task = loop.create_task(client_coro)
        turn_task = loop.create_task(chat.send_prompt(prompt))

        asyncio.run_coroutine_threadsafe(
            asyncio.wait({client_task, turn_task}, return_when=asyncio.FIRST_COMPLETED),
            loop,
        ).result()
    else:
        chat = Chat(
            chat_uuid,
            parent_event_uuid,
            settings.base_url,
            gql_client,
            model,
            strategy,
            autorun,
            LiveView(),
        )

        client_fut = asyncio.run_coroutine_threadsafe(client_coro, loop)
        input_handler = InputHandler()
        shell = Shell(prompt, loop, input_handler, chat, connection_tracker)
        shell.run()
        client_fut.cancel()

    print("Jump back into this chat by running:")
    print(f"  exponent shell --chat-id {chat_uuid}")
    print()
    print("Or continue in a web browser at:")
    print(f"  {chat.url()}")
    print()
    print("Bye!")


async def get_parent_event_uuid(
    gql_client: GraphQLClient, chat_uuid: str
) -> str | None:
    result = await gql_client.query(EVENTS_FOR_CHAT_QUERY, {"chatUuid": chat_uuid})
    events = result.get("eventsForChat", {}).get("events", [])

    if len(events) > 0:
        uuid = events[-1]["eventUuid"]
        assert isinstance(uuid, str)

        return uuid

    return None


def pause_spinner(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(self: "LiveView", *args: Any, **kwargs: Any) -> Any:
        self.spinner.hide()
        self.spinner = self.default_spinner
        result = func(self, *args, **kwargs)
        self.spinner.show()
        return result

    return wrapper


def stop_spinner(func: Callable[..., Any]) -> Callable[..., Any]:
    def wrapper(self: "LiveView", *args: Any, **kwargs: Any) -> Any:
        self.spinner.hide()
        self.spinner = self.default_spinner
        result = func(self, *args, **kwargs)
        return result

    return wrapper


class BaseView:
    def _render_block_header(self, text: str) -> list[str]:
        return [
            block_header_bg_seq(),
            block_header_fg_seq(),
            erase_line_seq(),
            " " + text,
            reset_attrs_seq(),
            "\n",
        ]

    def _render_block_padding(self) -> list[str]:
        return [
            block_body_bg_seq(),
            erase_line_seq(),
            reset_attrs_seq(),
            "\n",
        ]

    def _render_block_footer(self, text: str, status: str | None = None) -> list[str]:
        seqs: list[Any] = [
            block_header_bg_seq(),
            block_header_fg_seq(),
            erase_line_seq(),
            " ",
        ]

        if status == "completed":
            seqs.append(
                [
                    fg_color_seq(2),
                    bold_seq(),
                    "✓ ",
                    not_bold_seq(),
                    block_header_fg_seq(),
                ]
            )
        elif status == "rejected":
            seqs.append(
                [
                    fg_color_seq(1),
                    bold_seq(),
                    "𐄂 ",
                    not_bold_seq(),
                    block_header_fg_seq(),
                ]
            )
        elif status == "running":
            seqs.append(["⚙️ "])

        seqs.append([text, reset_attrs_seq(), "\n"])

        return seqs

    def _render_natural_edit_diff(self, write: dict[str, Any]) -> list[Any]:
        diff = generate_diff(write.get("originalFile") or "", write["newFile"]).strip()

        highlighted_diff = highlight_code(
            diff, "udiff", line_numbers=False, padding=(0, 2)
        )

        return [
            self._render_block_header("diff"),
            self._render_block_padding(),
            highlighted_diff,
            self._render_block_padding(),
        ]

    def _render_natural_edit_error(self, error: str) -> list[str]:
        return [
            fg_color_seq(1),
            f"\nError: {error.strip()}\n\n",
            reset_attrs_seq(),
        ]

    def _render_code_block_start(self, lang: str) -> list[Any]:
        return [
            self._render_block_header(lang),
            self._render_block_padding(),
        ]

    def _render_code_block_content(self, content: str, lang: str) -> str:
        return highlight_code(content.strip(), lang, line_numbers=False, padding=(0, 2))

    def _render_code_block_output(self, content: str) -> list[Any]:
        lines = pad_left(content, "  ").split("\n")
        output = [[block_body_bg_seq(), erase_line_seq(), line, "\n"] for line in lines]

        return [
            self._render_block_header("output"),
            self._render_block_padding(),
            output,
            self._render_block_padding(),
        ]

    def _render_file_write_block_start(self, path: str) -> list[Any]:
        return [
            self._render_block_header(f"Editing file {path}"),
            self._render_block_padding(),
        ]

    def _render_file_write_block_content(self, content: str, lang: str) -> str:
        return highlight_code(content.strip(), lang, line_numbers=False, padding=(0, 2))

    def _render_file_write_block_result(self, result: str | None) -> list[Any]:
        result = result or "Edit applied"
        return [self._render_block_footer(f"{result}", status="completed")]

    def _render_command_block_start(self, type_: str) -> list[Any]:
        return [
            self._render_block_header(type_.lower()),
            self._render_block_padding(),
        ]

    def _render_command_block_content(self, data: dict[str, Any]) -> list[Any]:
        lines = pad_left(data["filePath"].strip(), "  ").split("\n")
        return [[block_body_bg_seq(), erase_line_seq(), line, "\n"] for line in lines]

    def _render_command_block_output(
        self, content: str, type_: str, lang: str
    ) -> list[Any]:
        highlighted_result = highlight_code(
            content.strip(),
            lang,
            line_numbers=False,
            padding=(0, 2),
        )

        return [
            self._render_block_header("output"),
            self._render_block_padding(),
            highlighted_result,
            self._render_block_padding(),
            self._render_block_footer(
                f"{type_.lower()} command executed", status="completed"
            ),
        ]


class LiveView(BaseView):
    def __init__(self) -> None:
        self.buffer: Buffer = NullBuffer()
        self.command_data: dict[str, Any] | None = None
        self.confirmation_required: bool = True
        self.default_spinner = Spinner("Exponent is working...")
        self.msg_gen_spinner = Spinner("Exponent is thinking real hard...")
        self.exec_spinner = Spinner(
            "Exponent is waiting for the code to finish running..."
        )
        self.diff_spinner = Spinner("Exponent is generating file diff...")
        self.spinner = self.default_spinner

    def render_event(self, kind: str, event: dict[str, Any]) -> None:  # noqa: PLR0912
        if kind == "MessageStart":
            self._handle_message_start_event(event)
        elif kind == "MessageChunkEvent":
            if event["role"] == "assistant":
                self._handle_message_chunk_event(event)
        elif kind == "MessageEvent":
            if event["role"] == "assistant":
                self._handle_message_event(event)
        elif kind == "FileWriteChunkEvent":
            self._handle_file_write_chunk_event(event)
        elif kind == "FileWriteEvent":
            self._handle_file_write_event(event)
        elif kind == "FileWriteConfirmationEvent":
            self._handle_file_write_confirmation_event(event)
        elif kind == "FileWriteStartEvent":
            self._handle_file_write_start_event(event)
        elif kind == "FileWriteResultEvent":
            self._handle_file_write_result_event(event)
        elif kind == "CodeBlockChunkEvent":
            self._handle_code_block_chunk_event(event)
        elif kind == "CodeBlockEvent":
            self._handle_code_block_event(event)
        elif kind == "CodeBlockConfirmationEvent":
            self._handle_code_block_confirmation_event(event)
        elif kind == "CodeExecutionStartEvent":
            self._handle_code_execution_start_event(event)
        elif kind == "CodeExecutionEvent":
            self._handle_code_execution_event(event)
        elif kind == "CommandChunkEvent":
            if event["data"]["type"] in ["FILE_READ", "FILE_OPEN"]:
                self._handle_command_chunk_event(event)
        elif kind == "CommandEvent":
            if event["data"]["type"] in ["FILE_READ", "FILE_OPEN"]:
                self._handle_command_event(event)
        elif kind == "CommandConfirmationEvent":
            self._handle_command_confirmation_event(event)
        elif kind == "CommandStartEvent":
            self._handle_command_start_event(event)
        elif kind == "CommandResultEvent":
            self._handle_command_result_event(event)
        elif kind == "Error":
            raise ExponentError(event["message"])

    def start_turn(self) -> None:
        self.spinner.show()

    def end_turn(self) -> None:
        self.spinner.hide()

    @pause_spinner
    def _handle_message_start_event(self, event: dict[str, Any]) -> None:
        self.spinner = self.msg_gen_spinner

    @stop_spinner
    def _handle_message_chunk_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = CharBuffer(event_uuid)

        assert isinstance(self.buffer, CharBuffer)

        seqs.append(self.buffer.render_new_chars(event["content"]))
        render(seqs)

    @pause_spinner
    def _handle_message_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = CharBuffer(event_uuid)

        assert isinstance(self.buffer, CharBuffer)

        seqs.append(self.buffer.render_new_chars(event["content"]))
        seqs.append(["\n\n"])
        render(seqs)

    @stop_spinner
    def _handle_file_write_chunk_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = LineBuffer(event_uuid)

            seqs.append(
                [
                    self._render_block_header(f"Editing file {event["filePath"]}"),
                    self._render_block_padding(),
                ]
            )

        assert isinstance(self.buffer, LineBuffer)

        write = event["writeContent"]
        content = write.get("naturalEdit") or write.get("content") or event["content"]

        formatted_content = self._render_file_write_block_content(
            content, event["language"]
        )

        seqs.append(self.buffer.render_new_lines(formatted_content))

        if "newFile" in write and write["newFile"] is None:  # natural edit
            # when newFile is present in writeContent dict but is None
            # it indicates the server is generating original/new file pair

            seqs.append([self._render_block_padding(), "\n"])
            render(seqs)
            self.spinner = self.diff_spinner
            self.spinner.show()
        else:
            render(seqs)

    @pause_spinner
    def _handle_file_write_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = LineBuffer(event_uuid)
            seqs.append(self._render_file_write_block_start(event["filePath"]))

        assert isinstance(self.buffer, LineBuffer)

        write = event["writeContent"]

        if "newFile" in write:  # natural edit
            seqs.append(
                [
                    "\r",
                    move_cursor_up_seq(2),
                    erase_display_seq(),
                ]
            )

        content = write.get("naturalEdit") or write.get("content") or event["content"]

        formatted_content = self._render_file_write_block_content(
            content, event["language"]
        )

        seqs.append(
            [
                self.buffer.render_new_lines(formatted_content),
                self._render_block_padding(),
            ]
        )

        error = write.get("errorContent")

        if error is None:
            if write.get("newFile") is not None:
                seqs.append(self._render_natural_edit_diff(write))

            if event["requireConfirmation"]:
                seqs.append(
                    [
                        self._render_block_footer(
                            "Confirm edit with <C+y>. Sending a new message will dismiss code changes."
                        ),
                        "\n",
                    ]
                )
        else:
            seqs.append(self._render_natural_edit_error(error))

        render(seqs)
        self.confirmation_required = event["requireConfirmation"]

    @pause_spinner
    def _handle_file_write_confirmation_event(self, event: dict[str, Any]) -> None:
        if self.buffer.event_uuid != event["fileWriteUuid"]:
            return

        seqs = []

        if event["accepted"]:
            seqs.append(
                [
                    move_cursor_up_seq(2),
                    self._render_block_footer("Applying edit...", status="running"),
                    "\n",
                ]
            )
        elif self.confirmation_required:
            # user entered new prompt, cursor moved down
            # therefore we need to move it up, redraw status, and move it
            # back where it was

            seqs.append(
                [
                    move_cursor_up_seq(4),
                    self._render_block_footer("Edit dismissed", status="rejected"),
                    "\n\n\n",
                ]
            )

        render(seqs)

    @pause_spinner
    def _handle_file_write_start_event(self, event: dict[str, Any]) -> None:
        return

    @pause_spinner
    def _handle_file_write_result_event(self, event: dict[str, Any]) -> None:
        if self.buffer.event_uuid != event["fileWriteUuid"]:
            return

        seqs: list[Any] = []

        if self.confirmation_required:
            seqs.append([move_cursor_up_seq(2)])

        seqs.append([self._render_file_write_block_result(event["content"]), "\n"])

        render(seqs)

    @stop_spinner
    def _handle_command_chunk_event(self, event: dict[str, Any]) -> None:
        type_ = event["data"]["type"].lower()
        event_uuid = event["eventUuid"]
        seqs = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = CommandBuffer(event_uuid)

            seqs.append(
                [
                    self._render_block_header(type_),
                    self._render_block_padding(),
                ]
            )

        assert isinstance(self.buffer, CommandBuffer)

        formatted_content = self._render_command_block_content(event["data"])
        seqs.append(self.buffer.render_new_lines(formatted_content))
        render(seqs)

    @pause_spinner
    def _handle_command_event(self, event: dict[str, Any]) -> None:
        type_ = event["data"]["type"].lower()
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = CommandBuffer(event_uuid)
            seqs.append(self._render_command_block_start(event["data"]["type"]))

        assert isinstance(self.buffer, CommandBuffer)

        formatted_content = self._render_command_block_content(event["data"])
        seqs.append(self.buffer.render_new_lines(formatted_content))
        seqs.append(self._render_block_padding())

        if event["requireConfirmation"]:
            seqs.append(
                [
                    self._render_block_footer(
                        f"Confirm {type_} command with <C+y>. "
                        "Sending a new message will cancel this command."
                    ),
                    "\n",
                ]
            )

        render(seqs)
        self.command_data = event["data"]
        self.confirmation_required = event["requireConfirmation"]

    @pause_spinner
    def _handle_command_confirmation_event(self, event: dict[str, Any]) -> None:
        if self.buffer.event_uuid != event["commandUuid"]:
            return

        seqs = []

        if event["accepted"]:
            seqs.append(
                [
                    move_cursor_up_seq(2),
                    self._render_block_footer("Executing command...", status="running"),
                    "\n",
                ]
            )
        else:
            # user entered new prompt, cursor moved down
            # therefore we need to move it up, redraw status, and move it
            # back where it was

            seqs.append(
                [
                    move_cursor_up_seq(4),
                    self._render_block_footer(
                        "Command did not execute", status="rejected"
                    ),
                    "\n\n\n",
                ]
            )

        render(seqs)

    def _handle_command_start_event(self, event: dict[str, Any]) -> None:
        return

    @pause_spinner
    def _handle_command_result_event(self, event: dict[str, Any]) -> None:
        if self.buffer.event_uuid != event["commandUuid"] or self.command_data is None:
            return

        seqs: list[Any] = []

        if self.confirmation_required:
            seqs.append([move_cursor_up_seq(2)])

        seqs.append(
            [
                self._render_command_block_output(
                    event["content"],
                    self.command_data["type"],
                    self.command_data["language"],
                ),
                "\n",
            ]
        )

        render(seqs)
        self.command_data = None

    @stop_spinner
    def _handle_code_block_chunk_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = LineBuffer(event_uuid)

            seqs.append(
                [
                    self._render_block_header(event["language"]),
                    self._render_block_padding(),
                ]
            )

        assert isinstance(self.buffer, LineBuffer)

        formatted_content = self._render_code_block_content(
            event["content"], event["language"]
        )

        seqs.append(self.buffer.render_new_lines(formatted_content))
        render(seqs)

    @pause_spinner
    def _handle_code_block_event(self, event: dict[str, Any]) -> None:
        event_uuid = event["eventUuid"]
        seqs: list[Any] = []

        if self.buffer.event_uuid != event_uuid:
            self.buffer = LineBuffer(event_uuid)
            seqs.append(self._render_code_block_start(event["language"]))

        assert isinstance(self.buffer, LineBuffer)

        formatted_content = self._render_code_block_content(
            event["content"], event["language"]
        )

        seqs.append(self.buffer.render_new_lines(formatted_content))
        seqs.append(self._render_block_padding())

        if event["requireConfirmation"]:
            seqs.append(
                [
                    self._render_block_footer(
                        "Run this code now with <C+y>. Sending a new message will cancel this request."
                    ),
                    "\n",
                ]
            )

        render(seqs)
        self.confirmation_required = event["requireConfirmation"]

    @pause_spinner
    def _handle_code_block_confirmation_event(self, event: dict[str, Any]) -> None:
        if self.buffer.event_uuid != event["codeBlockUuid"]:
            return

        seqs = []

        if event["accepted"]:
            seqs.append(
                [
                    move_cursor_up_seq(2),
                    self._render_block_footer("Running code...", status="running"),
                    "\n",
                ]
            )
        else:
            # user entered new prompt, cursor moved down
            # therefore we need to move it up, redraw status, and move it
            # back where it was

            seqs.append(
                [
                    move_cursor_up_seq(4),
                    self._render_block_footer(
                        "Code did not execute", status="rejected"
                    ),
                    "\n\n\n",
                ]
            )

        render(seqs)

    @pause_spinner
    def _handle_code_execution_start_event(self, event: dict[str, Any]) -> None:
        self.spinner = self.exec_spinner

    @pause_spinner
    def _handle_code_execution_event(self, event: dict[str, Any]) -> None:
        if self.buffer.event_uuid != event["codeBlockUuid"]:
            return

        seqs: list[Any] = []

        if self.confirmation_required:
            seqs.append([move_cursor_up_seq(2)])

        seqs.append([self._render_code_block_output(event["content"]), "\n"])

        render(seqs)


class StaticView(BaseView):
    def __init__(self) -> None:
        self.command_data: dict[str, Any] | None = None

    def render_event(self, kind: str, event: dict[str, Any]) -> None:  # noqa: PLR0912
        if kind == "MessageEvent":
            if event["role"] == "assistant":
                print(event["content"].strip())
                print()

        elif kind == "FileWriteEvent":
            write = event["writeContent"]

            content = (
                write.get("naturalEdit") or write.get("content") or event["content"]
            )

            seqs = [
                self._render_file_write_block_start(event["filePath"]),
                self._render_file_write_block_content(content, event["language"]),
                self._render_block_padding(),
            ]

            error = write.get("errorContent")

            if error is None:
                if write.get("newFile") is not None:
                    seqs.append(self._render_natural_edit_diff(write))
            else:
                seqs.append(self._render_natural_edit_error(error))

            render(seqs)

        elif kind == "FileWriteResultEvent":
            seqs = [self._render_file_write_block_result(event["content"]), "\n"]
            render(seqs)

        elif kind == "CodeBlockEvent":
            seqs = [
                self._render_code_block_start(event["language"]),
                self._render_code_block_content(event["content"], event["language"]),
                self._render_block_padding(),
            ]

            render(seqs)

        elif kind == "CodeExecutionEvent":
            render([self._render_code_block_output(event["content"]), "\n"])

        elif kind == "CommandEvent":
            if event["data"]["type"] not in ["FILE_READ", "FILE_OPEN"]:
                return

            seqs = [
                self._render_command_block_start(event["data"]["type"]),
                self._render_command_block_content(event["data"]),
                self._render_block_padding(),
            ]

            render(seqs)
            self.command_data = event["data"]

        elif kind == "CommandResultEvent":
            if self.command_data is not None:
                seqs = [
                    self._render_command_block_output(
                        event["content"],
                        self.command_data["type"],
                        self.command_data["language"],
                    ),
                    "\n",
                ]

                render(seqs)
                self.command_data = None

        elif kind == "Error":
            print(f"Error: {event["message"]}")
            print()

    def start_turn(self) -> None:
        pass

    def end_turn(self) -> None:
        pass


class Chat:
    def __init__(
        self,
        chat_uuid: str,
        parent_event_uuid: str | None,
        base_url: str,
        gql_client: GraphQLClient,
        model: str,
        strategy: str,
        autorun: bool,
        view: StaticView | LiveView,
    ) -> None:
        self.chat_uuid = chat_uuid
        self.base_url = base_url
        self.gql_client = gql_client
        self.model = model
        self.strategy = strategy
        self.autorun = autorun
        self.view = view
        self.parent_uuid: str | None = parent_event_uuid
        self.block_row_offset = 0
        self.console = Console()
        self.code_block_uuid = None
        self.file_write_uuid = None
        self.command_uuid = None

    async def send_prompt(self, prompt: str) -> None:
        self.view.start_turn()

        await self.process_chat_subscription(
            {"prompt": {"message": prompt, "attachments": []}}
        )

    async def send_confirmation(self) -> None:
        if self.code_block_uuid is not None:
            await self.process_chat_subscription(
                {
                    "codeBlockConfirmation": {
                        "accepted": True,
                        "codeBlockUuid": self.code_block_uuid,
                    }
                }
            )
        elif self.file_write_uuid is not None:
            await self.process_chat_subscription(
                {
                    "fileWriteConfirmation": {
                        "accepted": True,
                        "fileWriteUuid": self.file_write_uuid,
                    }
                }
            )
        elif self.command_uuid is not None:
            await self.process_chat_subscription(
                {
                    "commandConfirmation": {
                        "accepted": True,
                        "commandUuid": self.command_uuid,
                    }
                }
            )

    def toggle_autorun(self) -> bool:
        self.autorun = not self.autorun
        return self.autorun

    async def halt_stream(self) -> None:
        await self.gql_client.query(
            HALT_CHAT_STREAM_MUTATION, {"chatUuid": self.chat_uuid}, "HaltChatStream"
        )

    def url(self) -> str:
        return f"{self.base_url}/chats/{self.chat_uuid}"

    async def process_chat_subscription(self, extra_vars: dict[str, Any]) -> None:
        vars = {
            "chatUuid": self.chat_uuid,
            "parentUuid": self.parent_uuid,
            "model": self.model,
            "strategyNameOverride": self.strategy,
            "useToolsConfig": "read_write",
            "requireConfirmation": not self.autorun,
        }

        vars.update(extra_vars)
        self.code_block_uuid = None
        self.file_write_uuid = None
        self.command_uuid = None

        try:
            async for response in self.gql_client.subscribe(
                CHAT_EVENTS_SUBSCRIPTION, vars
            ):
                event = response["authenticatedChat"]
                kind = event["__typename"]
                self.view.render_event(kind, event)
                self.parent_uuid = event.get("eventUuid") or self.parent_uuid

                if kind == "CodeBlockEvent" and event["requireConfirmation"]:
                    self.code_block_uuid = event["eventUuid"]
                elif kind == "FileWriteEvent" and event["requireConfirmation"]:
                    self.file_write_uuid = event["eventUuid"]
                elif kind == "CommandEvent" and event.get("requireConfirmation"):
                    self.command_uuid = event["eventUuid"]
        finally:
            self.view.end_turn()


class InputHandler:
    def __init__(self) -> None:
        self.kb = KeyBindings()
        self.shortcut_pressed = None
        self.session: PromptSession[Any] = PromptSession(
            completer=SlashCommandCompleter(),
            complete_while_typing=False,
            key_bindings=self.kb,
        )

        @self.kb.add("c-y")
        def _(event: Any) -> None:
            self.shortcut_pressed = "<c-y>"
            event.app.exit()

        @self.kb.add("c-d")
        def _(event: Any) -> None:
            self.shortcut_pressed = "<c-d>"
            event.app.exit()

    def prompt(self) -> str:
        self.shortcut_pressed = None
        user_input = self.session.prompt(HTML("<b><ansigreen>></ansigreen></b> "))

        if self.shortcut_pressed:
            return self.shortcut_pressed
        else:
            assert isinstance(user_input, str)
            return user_input


class Shell:
    def __init__(
        self,
        prompt: str | None,
        loop: asyncio.AbstractEventLoop,
        input_handler: InputHandler,
        chat: Chat,
        connection_tracker: ConnectionTracker,
    ) -> None:
        self.prompt = prompt
        self.loop = loop
        self.input_handler = input_handler
        self.chat = chat
        self.stream_fut: Future[Any] | None = None
        self.connection_tracker = connection_tracker

    def run(self) -> None:
        self._print_welcome_message()
        self._send_initial_prompt()

        while True:
            try:
                self._wait_for_stream_completion()
                text = self.input_handler.prompt()

                if text.startswith("/"):
                    self._run_command(text[1:].strip())
                elif text == "<c-y>":
                    self._confirm_execution()
                elif text in {"q", "exit"}:
                    print()
                    break
                elif text == "<c-d>":
                    print()
                    do_quit = self._ask_for_quit_confirmation()
                    print()

                    if do_quit:
                        break
                elif text:
                    print()
                    self._send_prompt(text)

            except KeyboardInterrupt:
                if self.stream_fut is not None:
                    self._run_coroutine(self.chat.halt_stream()).result()
                else:
                    do_quit = self._ask_for_quit_confirmation()
                    print()

                    if do_quit:
                        break

            except ExponentError as e:
                self._print_error_message(e)
                break

    def _ensure_connected(self) -> None:
        if not self.connection_tracker.is_connected():
            self._run_coroutine(self._wait_for_reconnection()).result()

    async def _wait_for_reconnection(self) -> None:
        render([clear_line_seq(), "Disconnected..."])
        await asyncio.sleep(1)
        spinner = Spinner("Reconnecting...")
        spinner.show()
        await self.connection_tracker.wait_for_reconnection()
        spinner.hide()
        render([fg_color_seq(2), bold_seq(), "✓ Reconnected"])
        await asyncio.sleep(1)
        render([clear_line_seq()])

    def _print_welcome_message(self) -> None:
        print("Welcome to ✨ \x1b[1;32mExponent \x1b[4:3mSHELL\x1b[0m ✨")
        print()
        print("Type 'q', 'exit' or press <C-c> to exit")
        print("Enter '/help' to see a list of available commands")
        print()

    def _print_error_message(self, e: ExponentError) -> None:
        print(f"\n\n\x1b[1;31m{e}\x1b[0m")
        print("\x1b[3;33m")
        print("Please try again and reach out if the problem persists.")
        print("\x1b[0m")

    def _show_help(self) -> None:
        print()

        for command, description in SLASH_COMMANDS.items():
            print(f"{command} - {description}")

        print()

    def _run_command(self, command: str) -> None:
        if command == "help":
            self._show_help()
        elif command == "autorun":
            self._toggle_autorun()
        elif command == "web":
            self._switch_chat_to_web()
        else:
            print(f"\nUnknown command: {command}\n")

    def _ask_for_quit_confirmation(self) -> bool:
        while True:
            answer = input("Do you want to quit Exponent shell? [y/N] ").strip().lower()

            if answer in {"y", "yes"}:
                return True
            elif answer in {"n", "no", ""}:
                return False

    def _toggle_autorun(self) -> None:
        if self.chat.toggle_autorun():
            print("\nAutorun mode enabled\n")
        else:
            print("\nAutorun mode disabled\n")

    def _switch_chat_to_web(self) -> None:
        url = self.chat.url()
        print(f"\nThis chat has been moved to {url}\n")

        if not inside_ssh_session():
            click.launch(url)

        while True:
            input()

    def _confirm_execution(self) -> None:
        render(
            [
                "\r",
                move_cursor_up_seq(1),
                clear_line_seq(),
            ]
        )

        self._ensure_connected()
        self.stream_fut = self._run_coroutine(self.chat.send_confirmation())

    def _send_initial_prompt(self) -> None:
        if self.prompt is not None:
            self._send_prompt(self.prompt)

    def _send_prompt(self, text: str) -> None:
        self._ensure_connected()
        self.stream_fut = self._run_coroutine(self.chat.send_prompt(text))

    def _wait_for_stream_completion(self) -> None:
        if self.stream_fut is not None:
            self.stream_fut.result()
            self.stream_fut = None

    def _run_coroutine(self, coro: Coroutine[Any, Any, Any]) -> Future[Any]:
        return asyncio.run_coroutine_threadsafe(coro, self.loop)


class SlashCommandCompleter(Completer):
    def get_completions(
        self, document: Document, complete_event: CompleteEvent
    ) -> Iterable[Completion]:
        text = document.text
        if text.startswith("/"):
            for command in SLASH_COMMANDS:
                if command.startswith(text):
                    yield Completion(command, start_position=-len(text))


class Buffer:
    def __init__(self, event_uuid: str) -> None:
        self.event_uuid = event_uuid


class NullBuffer(Buffer):
    def __init__(self) -> None:
        super().__init__("")


class CharBuffer(Buffer):
    def __init__(self, event_uuid: str) -> None:
        super().__init__(event_uuid)
        self.content_len = 0

    def render_new_chars(self, message: str) -> list[Any]:
        message = message.strip()
        chunk = message[self.content_len :]
        self.content_len = len(message)

        return [chunk]


class LineBuffer(Buffer):
    def __init__(self, event_uuid: str) -> None:
        super().__init__(event_uuid)
        self.line_count = 0

    def render_new_lines(
        self,
        code: str,
    ) -> list[Any]:
        seqs: list[Any] = []
        lines = code.split("\n")
        lines = lines[0 : len(lines) - 1]
        new_line_count = len(lines)

        if self.line_count > 0:
            seqs.append([move_cursor_up_seq(1), "\r"])
            lines = lines[self.line_count - 1 :]

        lines = [line + "\n" for line in lines]
        seqs.append(lines)
        self.line_count = new_line_count

        return seqs


# TODO maybe unify with LineBuffer?
class CommandBuffer(Buffer):
    def __init__(self, event_uuid: str) -> None:
        super().__init__(event_uuid)
        self.line_count = 0

    def render_new_lines(self, paths: list[Any]) -> list[Any]:
        seqs: list[Any] = []
        new_line_count = len(paths)

        if self.line_count > 0:
            seqs.append([move_cursor_up_seq(1), "\r"])
            paths = paths[self.line_count - 1 :]

        seqs.append(paths)
        self.line_count = new_line_count

        return seqs


def highlight_code(
    code: str, lang: str, line_numbers: bool = True, padding: tuple[int, int] = (0, 0)
) -> str:
    syntax = Syntax(
        code,
        lang,
        theme="monokai",
        line_numbers=line_numbers,
        word_wrap=True,
        padding=padding,
    )

    console = Console()

    with console.capture() as capture:
        console.print(syntax)

    return capture.get()


def generate_diff(before_lines: str, after_lines: str) -> str:
    diff = difflib.unified_diff(
        before_lines.split("\n"),
        after_lines.split("\n"),
        fromfile="before",
        tofile="after",
        lineterm="",
    )

    return "\n".join(list(diff)[2:])


def pad_left(text: str, padding: str) -> str:
    return "\n".join([padding + line for line in text.strip().split("\n")])


def fg_color_seq(c: int) -> str:
    return f"\x1b[{30 + c}m"


def bold_seq() -> str:
    return "\x1b[1m"


def not_bold_seq() -> str:
    return "\x1b[22m"


def block_header_bg_seq() -> str:
    return "\x1b[48;2;29;30;24m"


def block_header_fg_seq() -> str:
    return "\x1b[38;5;246m"


def block_body_bg_seq() -> str:
    return "\x1b[48;5;235m"


def erase_line_seq() -> str:
    return "\x1b[2K"


def erase_display_seq() -> str:
    return "\x1b[0J"


def reset_attrs_seq() -> str:
    return "\x1b[0m"


def clear_line_seq() -> str:
    return f"\r{reset_attrs_seq()}{erase_line_seq()}"


def move_cursor_up_seq(n: int) -> str:
    if n > 0:
        return f"\x1b[{n}A"
    else:
        return ""


def move_cursor_down_seq(n: int) -> str:
    if n > 0:
        return f"\x1b[{n}B"
    else:
        return ""


def render(seqs: str | list[Any] | None) -> None:
    print(collect(seqs), end="")
    sys.stdout.flush()


def collect(seqs: str | list[Any] | None) -> str:
    if seqs is None:
        return ""

    if isinstance(seqs, str):
        return seqs

    assert isinstance(seqs, list)

    text = ""

    for seq in seqs:
        text += collect(seq)

    return text



from exponent.core.remote_execution.languages.python import execute_python
from exponent.core.remote_execution.languages.shell import execute_shell
from exponent.core.remote_execution.session import RemoteExecutionClientSession
from exponent.core.remote_execution.types import (
    CodeExecutionRequest,
    CodeExecutionResponse,
)
from exponent.core.remote_execution.utils import assert_unreachable


async def execute_code(
    request: CodeExecutionRequest,
    session: RemoteExecutionClientSession,
    working_directory: str,
) -> CodeExecutionResponse:
    try:
        if request.language == "python":
            result = await execute_python(request.content, session.kernel)
        elif request.language == "shell":
            result = await execute_shell(
                request.content, working_directory, request.timeout
            )
        else:
            return assert_unreachable(request.language)
        if not result:
            result = "<No output>"
        return CodeExecutionResponse(
            content=result,
            correlation_id=request.correlation_id,
        )
    except Exception as e:  # noqa: BLE001 - TODO (Josh): Specialize errors for execution
        return CodeExecutionResponse(
            content="An error occurred while executing the code: " + str(e),
            correlation_id=request.correlation_id,
        )
